import pytest
from unittest.mock import call

from asynctest import TestCase as AsyncTestCase, mock as async_mock, call

from ...config.injection_context import InjectionContext
from ...utils.classloader import ClassLoader, ModuleLoadError

from ..plugin_registry import PluginRegistry
from ..protocol_registry import ProtocolRegistry

from ..error import ProtocolDefinitionValidationError


class TestPluginRegistry(AsyncTestCase):
    def setUp(self):
        self.registry = PluginRegistry()

        self.context = InjectionContext(enforce_typing=False)
        self.proto_registry = async_mock.MagicMock(
            register_message_types=async_mock.MagicMock(),
            register_controllers=async_mock.MagicMock(),
        )
        self.context.injector.bind_instance(ProtocolRegistry, self.proto_registry)

    async def test_setup(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name
        ctx = async_mock.MagicMock()
        self.registry._plugins[mod_name] = mod
        assert list(self.registry.plugin_names) == [mod_name]
        assert list(self.registry.plugins) == [mod]
        mod.setup = async_mock.CoroutineMock()
        await self.registry.init_context(ctx)
        mod.setup.assert_awaited_once_with(ctx)

    async def test_register_routes(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name
        app = async_mock.MagicMock()
        self.registry._plugins[mod_name] = mod
        mod.routes.register = async_mock.CoroutineMock()
        definition = async_mock.MagicMock()
        definition.versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v1_0",
            }
        ]

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[definition, mod.routes]),
        ) as load_module:
            await self.registry.register_admin_routes(app)

            calls = [
                call("definition", mod_name),
                call(f"{mod_name}.{definition.versions[0]['path']}.routes"),
            ]
            load_module.assert_has_calls(calls)
        assert mod.routes.register.call_count == 1

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[definition, ModuleLoadError()]),
        ) as load_module:
            await self.registry.register_admin_routes(app)

            calls = [
                call("definition", mod_name),
                call(f"{mod_name}.{definition.versions[0]['path']}.routes"),
            ]
            load_module.assert_has_calls(calls)
        assert mod.routes.register.call_count == 1

    async def test_register_routes_mod_no_version(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name
        app = async_mock.MagicMock()
        self.registry._plugins[mod_name] = mod
        mod.routes.register = async_mock.CoroutineMock()

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[None, mod.routes]),
        ) as load_module:
            await self.registry.register_admin_routes(app)

            calls = [call("definition", mod_name), call(f"{mod_name}.routes")]
            load_module.assert_has_calls(calls)
        assert mod.routes.register.call_count == 1

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[None, ModuleLoadError()]),
        ) as load_module:
            await self.registry.register_admin_routes(app)

            calls = [
                call("definition", mod_name),
                call(f"{mod_name}.routes"),
            ]
            load_module.assert_has_calls(calls)
        assert mod.routes.register.call_count == 1

    async def test_post_process_routes(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name
        app = async_mock.MagicMock()
        self.registry._plugins[mod_name] = mod
        mod.routes.post_process_routes = async_mock.MagicMock()
        definition = async_mock.MagicMock()
        definition.versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v1_0",
            }
        ]

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[definition, mod.routes]),
        ) as load_module:
            self.registry.post_process_routes(app)

            calls = [
                call("definition", mod_name),
                call(f"{mod_name}.{definition.versions[0]['path']}.routes"),
            ]
            load_module.assert_has_calls(calls)
        assert mod.routes.post_process_routes.call_count == 1

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[definition, ModuleLoadError()]),
        ) as load_module:
            self.registry.post_process_routes(app)

            calls = [
                call("definition", mod_name),
                call(f"{mod_name}.{definition.versions[0]['path']}.routes"),
            ]
            load_module.assert_has_calls(calls)
        assert mod.routes.post_process_routes.call_count == 1

    async def test_post_process_routes_mod_no_version(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name
        app = async_mock.MagicMock()
        self.registry._plugins[mod_name] = mod
        mod.routes.register = async_mock.CoroutineMock()

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[None, mod.routes]),
        ) as load_module:
            self.registry.post_process_routes(app)

            calls = [call("definition", mod_name), call(f"{mod_name}.routes")]
            load_module.assert_has_calls(calls)
        assert mod.routes.post_process_routes.call_count == 1

        with async_mock.patch.object(
            ClassLoader,
            "load_module",
            async_mock.MagicMock(side_effect=[None, ModuleLoadError()]),
        ) as load_module:
            self.registry.post_process_routes(app)

            calls = [call("definition", mod_name), call(f"{mod_name}.routes")]
            load_module.assert_has_calls(calls)
        assert mod.routes.post_process_routes.call_count == 1

    async def test_validate_version_not_a_list(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions_not_a_list = {}

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions_not_a_list, mod_name)

    async def test_validate_version_list_element_not_an_object(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [{}, []]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_validate_version_list_element_empty(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = []

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_validate_version_list_missing_attribute(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                # "path": "v1_0", # missing
            }
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_validate_version_negative_version(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": -1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v1_0",
            }
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_validate_version_min_greater_current(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 1,
                "current_minor_version": 0,
                "path": "v1_0",
            }
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_validate_version_multiple_major(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v1_0",
            },
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 1,
                "path": "v1_1",
            },
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_validate_version_bad_path(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 1,
                "current_minor_version": 0,
                "path": "v1_0",
            }
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock(return_value=None)
        ) as load_module:
            with pytest.raises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_validate_version_list_correct(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v1_0",
            },
            {
                "major_version": 2,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v2_0",
            },
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            assert self.registry.validate_version(versions, mod_name) is True

            load_module.has_calls(
                call(versions[0]["path"], mod_name),
                call(versions[1]["path"], mod_name),
            )

    async def test_validate_version_list_extra_attributes_ok(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v1_0",
                "not": "an attribute",
            }
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            assert self.registry.validate_version(versions, mod_name) is True

    async def test_validate_version_no_such_mod(self):
        mod_name = "no_mod"
        mod = async_mock.MagicMock()
        mod.__name__ = mod_name

        versions = [
            {
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 0,
                "path": "v1_0",
            }
        ]

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.return_value = None

            with self.assertRaises(ProtocolDefinitionValidationError):
                self.registry.validate_version(versions, mod_name)

    async def test_register_plugin_already_present(self):
        mod_name = "test_mod"
        mod = async_mock.MagicMock()
        self.registry._plugins[mod_name] = mod
        assert mod == self.registry.register_plugin(mod_name)

    async def test_register_plugin_load_x(self):
        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = ModuleLoadError("failure to load")
            assert self.registry.register_plugin("dummy") is None

    async def test_register_plugin_no_mod(self):
        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.return_value = None
            assert self.registry.register_plugin("dummy") is None

    async def test_register_plugin_no_definition(self):
        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = [
                async_mock.MagicMock(),  # module
                None,  # routes
                None,  # message types
                None,  # definition
            ]
            assert self.registry.register_plugin("dummy") is None

    async def test_register_plugin_no_versions(self):
        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = [
                async_mock.MagicMock(),  # module
                None,  # routes
                None,  # message types
                "str-has-no-versions-attr",  # definition without versions attr
            ]
            assert self.registry.register_plugin("dummy") is None

    async def test_register_definitions_malformed(self):
        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = [
                async_mock.MagicMock(),  # module
                None,  # routes
                None,  # message types
                async_mock.MagicMock(versions="not-a-list"),
            ]
            assert self.registry.register_plugin("dummy") is None

    async def test_register_package_x(self):
        with async_mock.patch.object(
            ClassLoader, "scan_subpackages", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = ModuleLoadError()
            assert not self.registry.register_package("dummy")

    async def test_load_protocols_load_x(self):
        mock_plugin = async_mock.MagicMock(__name__="dummy")
        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = ModuleLoadError()
            await self.registry.load_protocols(None, mock_plugin)
            assert load_module.call_count == 1

    async def test_load_protocols_load_mod(self):
        mock_plugin = async_mock.MagicMock(__name__="dummy")
        mock_mod = async_mock.MagicMock()
        mock_mod.MESSAGE_TYPES = async_mock.MagicMock()
        mock_mod.CONTROLLERS = async_mock.MagicMock()

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.return_value = mock_mod
            await self.registry.load_protocols(self.context, mock_plugin)

    async def test_load_protocols_no_mod_load_x(self):
        mock_plugin = async_mock.MagicMock(__name__="dummy")

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = [None, ModuleLoadError()]
            await self.registry.load_protocols(self.context, mock_plugin)
            assert load_module.call_count == 2

    async def test_load_protocols_no_mod_def_no_message_types(self):
        mock_plugin = async_mock.MagicMock(__name__="dummy")
        mock_def = async_mock.MagicMock(
            versions=[
                {
                    "major_version": 1,
                    "minimum_minor_version": 0,
                    "current_minor_version": 0,
                    "path": "v1_0",
                }
            ]
        )

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = [None, mock_def, ModuleLoadError()]
            await self.registry.load_protocols(self.context, mock_plugin)
            assert load_module.call_count == 3

    async def test_load_protocols_no_mod_def_message_types(self):
        mock_plugin = async_mock.MagicMock(__name__="dummy")
        mock_def = async_mock.MagicMock(
            versions=[
                {
                    "major_version": 1,
                    "minimum_minor_version": 0,
                    "current_minor_version": 0,
                    "path": "v1_0",
                },
                {
                    "major_version": 2,
                    "minimum_minor_version": 0,
                    "current_minor_version": 0,
                    "path": "v2_0",
                },
            ]
        )
        mock_mod = async_mock.MagicMock()
        mock_mod.MESSAGE_TYPES = async_mock.MagicMock()
        mock_mod.CONTROLLERS = async_mock.MagicMock()

        with async_mock.patch.object(
            ClassLoader, "load_module", async_mock.MagicMock()
        ) as load_module:
            load_module.side_effect = [None, mock_def, mock_mod, mock_mod]
            await self.registry.load_protocols(self.context, mock_plugin)
            assert load_module.call_count == 4

    def test_repr(self):
        assert type(repr(self.registry)) is str



from asynctest import TestCase as AsyncTestCase

from ....core.in_memory import InMemoryProfile
from ....protocols.connections.v1_0.messages.connection_invitation import (
    ConnectionInvitation,
)
from ....protocols.connections.v1_0.messages.connection_request import ConnectionRequest
from ....protocols.connections.v1_0.models.connection_detail import ConnectionDetail
from ....storage.base import BaseStorage
from ....storage.error import StorageNotFoundError

from ..conn_record import ConnRecord
from ..diddoc.diddoc import DIDDoc


class TestConnRecord(AsyncTestCase):
    def setUp(self):
        self.session = InMemoryProfile.test_session()

        self.test_seed = "testseed000000000000000000000001"
        self.test_did = "55GkHamhTU1ZbTbV2ab9DE"
        self.test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
        self.test_endpoint = "http://localhost"

        self.test_target_did = "GbuDUYXaUZRfHD2jeDuQuP"
        self.test_target_verkey = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"

        self.test_conn_record = ConnRecord(
            my_did=self.test_did,
            their_did=self.test_target_did,
            their_role=ConnRecord.Role.REQUESTER.rfc23,
            state=ConnRecord.State.COMPLETED.rfc23,
        )
        assert self.test_conn_record.their_role == ConnRecord.Role.REQUESTER.rfc160
        assert self.test_conn_record.state == ConnRecord.State.COMPLETED.rfc160
        assert self.test_conn_record.rfc23_state == ConnRecord.State.COMPLETED.rfc23

    async def test_get_enums(self):
        assert ConnRecord.Role.get("Larry") is None
        assert ConnRecord.State.get("a suffusion of yellow") is None

        assert (
            ConnRecord.Role.get(ConnRecord.Role.REQUESTER) is ConnRecord.Role.REQUESTER
        )

        assert (
            ConnRecord.State.get(ConnRecord.State.RESPONSE) is ConnRecord.State.RESPONSE
        )

        assert ConnRecord.Role.REQUESTER.flip() is ConnRecord.Role.RESPONDER
        assert ConnRecord.Role.get(
            ConnRecord.Role.REQUESTER.rfc160
        ) is ConnRecord.Role.get(ConnRecord.Role.REQUESTER.rfc23)
        assert ConnRecord.Role.REQUESTER == ConnRecord.Role.REQUESTER.rfc160  # check ==
        assert ConnRecord.Role.REQUESTER == ConnRecord.Role.REQUESTER.rfc23
        assert ConnRecord.Role.REQUESTER != ConnRecord.Role.RESPONDER.rfc23

    async def test_state_rfc23strict(self):
        for state in (
            ConnRecord.State.INIT,
            ConnRecord.State.ABANDONED,
            ConnRecord.State.COMPLETED,
        ):
            assert state.rfc23strict(their_role=None) == state.value[1]

        for state in (ConnRecord.State.INVITATION, ConnRecord.State.RESPONSE):
            assert (
                state.rfc23strict(their_role=ConnRecord.Role.REQUESTER)
                == f"{state.value[1]}-sent"
            )
            assert (
                state.rfc23strict(their_role=ConnRecord.Role.RESPONDER)
                == f"{state.value[1]}-received"
            )

        assert (
            ConnRecord.State.REQUEST.rfc23strict(their_role=ConnRecord.Role.REQUESTER)
            == f"{ConnRecord.State.REQUEST.value[1]}-received"
        )
        assert (
            ConnRecord.State.REQUEST.rfc23strict(their_role=ConnRecord.Role.RESPONDER)
            == f"{ConnRecord.State.REQUEST.value[1]}-sent"
        )

    async def test_save_retrieve_compare(self):
        record = ConnRecord(my_did=self.test_did)
        connection_id = await record.save(self.session)
        fetched = await ConnRecord.retrieve_by_id(self.session, connection_id)
        assert fetched and fetched == record

        bad_record = ConnRecord(my_did=None)
        bad_record._id = record._id
        bad_record.created_at = record.created_at
        bad_record.updated_at = record.updated_at
        assert bad_record != record

        record = ConnRecord(
            state=ConnRecord.State.INIT,  # exercise init State by enum
            my_did=self.test_did,
            their_role=ConnRecord.Role.REQUESTER,  # exercise init Role by enum
        )
        connection_id = await record.save(self.session)
        fetched = await ConnRecord.retrieve_by_id(self.session, connection_id)
        assert fetched and fetched == record
        assert fetched.state is ConnRecord.State.INIT.rfc160
        assert ConnRecord.State.get(fetched.state) is ConnRecord.State.INIT
        assert fetched.their_role is ConnRecord.Role.REQUESTER.rfc160
        assert ConnRecord.Role.get(fetched.their_role) is ConnRecord.Role.REQUESTER

        record160 = ConnRecord(
            state=ConnRecord.State.INIT.rfc23,
            my_did=self.test_did,
            their_role=ConnRecord.Role.REQUESTER.rfc23,
        )
        record160._id = record._id
        record160.created_at = record.created_at
        record160.updated_at = record.updated_at
        assert record160 == record

    async def test_retrieve_by_did(self):
        record = ConnRecord(
            my_did=self.test_did,
            their_did=self.test_target_did,
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            state=ConnRecord.State.COMPLETED.rfc23,
        )
        rec_id = await record.save(self.session)
        result = await ConnRecord.retrieve_by_did(
            session=self.session,
            my_did=self.test_did,
            their_did=self.test_target_did,
            their_role=ConnRecord.Role.RESPONDER.rfc160,
        )
        assert result == record

    async def test_from_storage_with_initiator_old(self):
        record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.COMPLETED)
        ser = record.serialize()
        ser["initiator"] = "self"  # old-style ConnectionRecord
        ConnRecord.from_storage("conn-id", ser)

    async def test_retrieve_by_invitation_key(self):
        record = ConnRecord(
            my_did=self.test_did,
            their_did=self.test_target_did,
            their_role=ConnRecord.Role.RESPONDER.rfc160,
            state=ConnRecord.State.INVITATION.rfc23,
            invitation_key="dummy",
        )
        await record.save(self.session)
        result = await ConnRecord.retrieve_by_invitation_key(
            session=self.session,
            invitation_key="dummy",
            their_role=ConnRecord.Role.RESPONDER.rfc23,
        )
        assert result == record
        with self.assertRaises(StorageNotFoundError):
            await ConnRecord.retrieve_by_invitation_key(
                session=self.session,
                invitation_key="dummy",
                their_role=ConnRecord.Role.REQUESTER.rfc23,
            )

    async def test_retrieve_by_request_id(self):
        record = ConnRecord(
            my_did=self.test_did,
            their_did=self.test_target_did,
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            state=ConnRecord.State.COMPLETED.rfc23,
            request_id="abc123",
        )
        await record.save(self.session)
        result = await ConnRecord.retrieve_by_request_id(
            session=self.session, request_id="abc123"
        )
        assert result == record

    async def test_completed_is_ready(self):
        record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.COMPLETED)
        connection_id = await record.save(self.session)
        fetched = await ConnRecord.retrieve_by_id(self.session, connection_id)

        assert fetched.is_ready == True

    async def test_response_is_ready(self):
        record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.RESPONSE)
        connection_id = await record.save(self.session)
        fetched = await ConnRecord.retrieve_by_id(self.session, connection_id)

        assert fetched.is_ready is True

    async def test_request_is_not_ready(self):
        record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.REQUEST)
        connection_id = await record.save(self.session)
        fetched = await ConnRecord.retrieve_by_id(self.session, connection_id)

        assert fetched.is_ready is False

    async def test_invitation_is_not_multi_use(self):
        record = ConnRecord(
            my_did=self.test_did,
            state=ConnRecord.State.INVITATION.rfc23,
            invitation_mode=ConnRecord.INVITATION_MODE_ONCE,
        )
        connection_id = await record.save(self.session)
        fetched = await ConnRecord.retrieve_by_id(self.session, connection_id)

        assert fetched.is_multiuse_invitation is False

    async def test_invitation_is_multi_use(self):
        record = ConnRecord(
            my_did=self.test_did,
            state=ConnRecord.State.INVITATION.rfc23,
            invitation_mode=ConnRecord.INVITATION_MODE_MULTI,
        )
        connection_id = await record.save(self.session)
        fetched = await ConnRecord.retrieve_by_id(self.session, connection_id)

        assert fetched.is_multiuse_invitation is True

    async def test_attach_retrieve_invitation(self):
        record = ConnRecord(
            my_did=self.test_did,
            state=ConnRecord.State.INVITATION.rfc23,
        )
        connection_id = await record.save(self.session)

        invi = ConnectionInvitation(
            label="abc123",
            recipient_keys=[self.test_verkey],
            endpoint="http://localhost:8999",
        )
        await record.attach_invitation(self.session, invi)
        retrieved = await record.retrieve_invitation(self.session)
        assert isinstance(retrieved, ConnectionInvitation)

    async def test_attach_retrieve_request(self):
        record = ConnRecord(
            my_did=self.test_did,
            state=ConnRecord.State.INVITATION.rfc23,
        )
        connection_id = await record.save(self.session)

        req = ConnectionRequest(
            connection=ConnectionDetail(
                did=self.test_did, did_doc=DIDDoc(self.test_did)
            ),
            label="abc123",
        )
        await record.attach_request(self.session, req)
        retrieved = await record.retrieve_request(self.session)
        assert isinstance(retrieved, ConnectionRequest)

    async def test_ser_rfc23_state_present(self):
        record = ConnRecord(
            state=ConnRecord.State.INVITATION,
            my_did=self.test_did,
            their_role=ConnRecord.Role.REQUESTER,
        )
        ser = record.serialize()
        assert ser["rfc23_state"] == f"{ConnRecord.State.INVITATION.value[1]}-sent"

    async def test_deser_old_style_record(self):
        record = ConnRecord(
            state=ConnRecord.State.INIT,
            my_did=self.test_did,
            their_role=ConnRecord.Role.REQUESTER,
        )
        ser = record.serialize()
        ser["initiator"] = "self"  # redundant vs. role as per RFC 160 or RFC 23
        deser = ConnRecord.deserialize(ser)
        reser = deser.serialize()
        assert "initiator" not in reser

    async def test_metadata_set_get(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        await record.metadata_set(self.session, "key", {"test": "value"})
        retrieved = await record.metadata_get(self.session, "key")
        assert retrieved == {"test": "value"}

    async def test_metadata_set_get_str(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        await record.metadata_set(self.session, "key", "value")
        retrieved = await record.metadata_get(self.session, "key")
        assert retrieved == "value"

    async def test_metadata_set_update_get(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        await record.metadata_set(self.session, "key", {"test": "value"})
        await record.metadata_set(self.session, "key", {"test": "updated"})
        retrieved = await record.metadata_get(self.session, "key")
        assert retrieved == {"test": "updated"}

    async def test_metadata_get_without_set_is_none(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        assert await record.metadata_get(self.session, "key") is None

    async def test_metadata_get_default(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        assert await record.metadata_get(self.session, "key", {"test": "default"}) == {
            "test": "default"
        }

    async def test_metadata_set_delete_get_is_none(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        await record.metadata_set(self.session, "key", {"test": "value"})
        await record.metadata_delete(self.session, "key")
        assert await record.metadata_get(self.session, "key") is None

    async def test_metadata_delete_without_set_raise_error(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        with self.assertRaises(KeyError) as exc:
            await record.metadata_delete(self.session, "key")
            assert "key not found in connection metadata" in exc.msg

    async def test_metadata_get_all(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        await record.metadata_set(self.session, "key", {"test": "value"})
        await record.metadata_set(self.session, "key", {"test": "updated"})
        await record.metadata_set(self.session, "other", {"test": "other"})
        retrieved = await record.metadata_get_all(self.session)
        assert retrieved == {"key": {"test": "updated"}, "other": {"test": "other"}}

    async def test_metadata_get_all_without_set_is_empty(self):
        record = ConnRecord(
            my_did=self.test_did,
        )
        await record.save(self.session)
        assert await record.metadata_get_all(self.session) == {}



"""Basic in-memory storage implementation (non-wallet)."""

from typing import Mapping, Sequence

from ..core.in_memory import InMemoryProfile

from .base import (
    DEFAULT_PAGE_SIZE,
    BaseStorage,
    BaseStorageSearch,
    BaseStorageSearchSession,
    validate_record,
)
from .error import (
    StorageDuplicateError,
    StorageNotFoundError,
    StorageSearchError,
)
from .record import StorageRecord


class InMemoryStorage(BaseStorage, BaseStorageSearch):
    """Basic in-memory storage class."""

    def __init__(self, profile: InMemoryProfile):
        """
        Initialize a `InMemoryStorage` instance.

        Args:
            profile: The in-memory profile instance

        """
        self.profile = profile

    async def add_record(self, record: StorageRecord):
        """
        Add a new record to the store.

        Args:
            record: `StorageRecord` to be stored

        Raises:
            StorageError: If no record is provided
            StorageError: If the record has no ID

        """
        validate_record(record)
        if record.id in self.profile.records:
            raise StorageDuplicateError("Duplicate record")
        self.profile.records[record.id] = record

    async def get_record(
        self, record_type: str, record_id: str, options: Mapping = None
    ) -> StorageRecord:
        """
        Fetch a record from the store by type and ID.

        Args:
            record_type: The record type
            record_id: The record id
            options: A dictionary of backend-specific options

        Returns:
            A `StorageRecord` instance

        Raises:
            StorageNotFoundError: If the record is not found

        """
        row = self.profile.records.get(record_id)
        if row and row.type == record_type:
            return row
        if not row:
            raise StorageNotFoundError("Record not found: {}".format(record_id))

    async def update_record(self, record: StorageRecord, value: str, tags: Mapping):
        """
        Update an existing stored record's value.

        Args:
            record: `StorageRecord` to update
            value: The new value
            tags: The new tags

        Raises:
            StorageNotFoundError: If record not found

        """
        validate_record(record)
        oldrec = self.profile.records.get(record.id)
        if not oldrec:
            raise StorageNotFoundError("Record not found: {}".format(record.id))
        self.profile.records[record.id] = oldrec._replace(value=value, tags=tags)

    async def delete_record(self, record: StorageRecord):
        """
        Delete a record.

        Args:
            record: `StorageRecord` to delete

        Raises:
            StorageNotFoundError: If record not found

        """
        validate_record(record, delete=True)
        if record.id not in self.profile.records:
            raise StorageNotFoundError("Record not found: {}".format(record.id))
        del self.profile.records[record.id]

    async def find_all_records(
        self,
        type_filter: str,
        tag_query: Mapping = None,
        options: Mapping = None,
    ):
        """Retrieve all records matching a particular type filter and tag query."""
        results = []
        for record in self.profile.records.values():
            if record.type == type_filter and tag_query_match(record.tags, tag_query):
                results.append(record)
        return results

    async def delete_all_records(
        self,
        type_filter: str,
        tag_query: Mapping = None,
    ):
        """Remove all records matching a particular type filter and tag query."""
        ids = []
        for record_id, record in self.profile.records.items():
            if record.type == type_filter and tag_query_match(record.tags, tag_query):
                ids.append(record_id)
        for record_id in ids:
            del self.profile.records[record_id]

    def search_records(
        self,
        type_filter: str,
        tag_query: Mapping = None,
        page_size: int = None,
        options: Mapping = None,
    ) -> "InMemoryStorageSearch":
        """
        Search stored records.

        Args:
            type_filter: Filter string
            tag_query: Tags to query
            page_size: Page size
            options: Dictionary of backend-specific options

        Returns:
            An instance of `InMemoryStorageSearch`

        """
        return InMemoryStorageSearch(
            self.profile, type_filter, tag_query, page_size, options
        )


def tag_value_match(value: str, match: dict) -> bool:
    """Match a single tag against a tag subquery.

    TODO: What type coercion is needed? (support int or float values?)
    """
    if len(match) != 1:
        raise StorageSearchError("Unsupported subquery: {}".format(match))
    if value is None:
        return False
    op = list(match.keys())[0]
    cmp_val = match[op]
    if op == "$in":
        if not isinstance(cmp_val, list):
            raise StorageSearchError("Expected list for $in value")
        chk = value in cmp_val
    else:
        if not isinstance(cmp_val, str):
            raise StorageSearchError("Expected string for filter value")
        if op == "$neq":
            chk = value != cmp_val
        elif op == "$gt":
            chk = float(value) > float(cmp_val)
        elif op == "$gte":
            chk = float(value) >= float(cmp_val)
        elif op == "$lt":
            chk = float(value) < float(cmp_val)
        elif op == "$lte":
            chk = float(value) <= float(cmp_val)
        # elif op == "$like":  NYI
        else:
            raise StorageSearchError(f"Unsupported match operator: {op}")
    return chk


def tag_query_match(tags: dict, tag_query: dict) -> bool:
    """Match simple tag filters (string values)."""
    result = True
    if not tags:
        tags = {}
    if tag_query:
        for k, v in tag_query.items():
            if k == "$or":
                if not isinstance(v, list):
                    raise StorageSearchError("Expected list for $or filter value")
                chk = False
                for opt in v:
                    if tag_query_match(tags, opt):
                        chk = True
                        break
            elif k == "$not":
                if not isinstance(v, dict):
                    raise StorageSearchError("Expected dict for $not filter value")
                chk = not tag_query_match(tags, v)
            elif k[0] == "$":
                raise StorageSearchError("Unexpected filter operator: {}".format(k))
            elif isinstance(v, str):
                chk = tags.get(k) == v
            elif isinstance(v, dict):
                chk = tag_value_match(tags.get(k), v)
            else:
                raise StorageSearchError(
                    "Expected string or dict for filter value, got {}".format(v)
                )
            if not chk:
                result = False
                break
    return result


class InMemoryStorageSearch(BaseStorageSearchSession):
    """Represent an active stored records search."""

    def __init__(
        self,
        profile: InMemoryProfile,
        type_filter: str,
        tag_query: Mapping,
        page_size: int = None,
        options: Mapping = None,
    ):
        """
        Initialize a `InMemoryStorageSearch` instance.

        Args:
            profile: The in-memory profile to search
            type_filter: Filter string
            tag_query: Tags to search
            page_size: Size of page to return
            options: Dictionary of backend-specific options

        """
        self._cache = profile.records.copy()
        self._iter = iter(self._cache)
        self.page_size = page_size or DEFAULT_PAGE_SIZE
        self.tag_query = tag_query
        self.type_filter = type_filter

    async def fetch(self, max_count: int = None) -> Sequence[StorageRecord]:
        """
        Fetch the next list of results from the store.

        Args:
            max_count: Max number of records to return. If not provided,
              defaults to the backend's preferred page size

        Returns:
            A list of `StorageRecord` instances

        Raises:
            StorageSearchError: If the search query has not been opened

        """
        if self._cache is None:
            raise StorageSearchError("Search query is complete")

        ret = []
        check_type = self.type_filter
        i = max_count or self.page_size

        while i > 0:
            try:
                id = next(self._iter)
            except StopIteration:
                break
            record = self._cache[id]
            if record.type == check_type and tag_query_match(
                record.tags, self.tag_query
            ):
                ret.append(record)
                i -= 1

        if not ret:
            self._cache = None

        return ret

    async def close(self):
        """Dispose of the search query."""
        self._cache = None



"""
The Dispatcher.

The dispatcher is responsible for coordinating data flow between handlers, providing
lifecycle hook callbacks storing state for message threads, etc.
"""

import asyncio
import logging
import os
from typing import Callable, Coroutine, Union

from aiohttp.web import HTTPException

from ..core.profile import Profile
from ..messaging.agent_message import AgentMessage
from ..messaging.error import MessageParseError
from ..messaging.models.base import BaseModelError
from ..messaging.request_context import RequestContext
from ..messaging.responder import BaseResponder
from ..messaging.util import datetime_now
from ..protocols.connections.v1_0.manager import ConnectionManager
from ..protocols.problem_report.v1_0.message import ProblemReport
from ..transport.inbound.message import InboundMessage
from ..transport.outbound.message import OutboundMessage
from ..utils.stats import Collector
from ..utils.task_queue import CompletedTask, PendingTask, TaskQueue
from ..utils.tracing import trace_event, get_timer

from .error import ProtocolMinorVersionNotSupported
from .protocol_registry import ProtocolRegistry

LOGGER = logging.getLogger(__name__)


class Dispatcher:
    """
    Dispatcher class.

    Class responsible for dispatching messages to message handlers and responding
    to other agents.
    """

    def __init__(self, profile: Profile):
        """Initialize an instance of Dispatcher."""
        self.collector: Collector = None
        self.profile = profile
        self.task_queue: TaskQueue = None

    async def setup(self):
        """Perform async instance setup."""
        self.collector = self.profile.inject(Collector, required=False)
        max_active = int(os.getenv("DISPATCHER_MAX_ACTIVE", 50))
        self.task_queue = TaskQueue(
            max_active=max_active, timed=bool(self.collector), trace_fn=self.log_task
        )

    def put_task(
        self, coro: Coroutine, complete: Callable = None, ident: str = None
    ) -> PendingTask:
        """Run a task in the task queue, potentially blocking other handlers."""
        return self.task_queue.put(coro, complete, ident)

    def run_task(
        self, coro: Coroutine, complete: Callable = None, ident: str = None
    ) -> asyncio.Task:
        """Run a task in the task queue, potentially blocking other handlers."""
        return self.task_queue.run(coro, complete, ident)

    def log_task(self, task: CompletedTask):
        """Log a completed task using the stats collector."""
        if task.exc_info and not issubclass(task.exc_info[0], HTTPException):
            # skip errors intentionally returned to HTTP clients
            LOGGER.exception(
                "Handler error: %s", task.ident or "", exc_info=task.exc_info
            )
        if self.collector:
            timing = task.timing
            if "queued" in timing:
                self.collector.log(
                    "Dispatcher:queued", timing["unqueued"] - timing["queued"]
                )
            if task.ident:
                self.collector.log(task.ident, timing["ended"] - timing["started"])

    def queue_message(
        self,
        profile: Profile,
        inbound_message: InboundMessage,
        send_outbound: Coroutine,
        send_webhook: Coroutine = None,
        complete: Callable = None,
    ) -> PendingTask:
        """
        Add a message to the processing queue for handling.

        Args:
            profile: The profile associated with the inbound message
            inbound_message: The inbound message instance
            send_outbound: Async function to send outbound messages
            send_webhook: Async function to dispatch a webhook
            complete: Function to call when the handler has completed

        Returns:
            A pending task instance resolving to the handler task

        """
        return self.put_task(
            self.handle_message(profile, inbound_message, send_outbound, send_webhook),
            complete,
        )

    async def handle_message(
        self,
        profile: Profile,
        inbound_message: InboundMessage,
        send_outbound: Coroutine,
        send_webhook: Coroutine = None,
    ):
        """
        Configure responder and message context and invoke the message handler.

        Args:
            profile: The profile associated with the inbound message
            inbound_message: The inbound message instance
            send_outbound: Async function to send outbound messages
            send_webhook: Async function to dispatch a webhook

        Returns:
            The response from the handler

        """
        r_time = get_timer()

        error_result = None
        try:
            message = await self.make_message(inbound_message.payload)
        except MessageParseError as e:
            LOGGER.error(f"Message parsing failed: {str(e)}, sending problem report")
            error_result = ProblemReport(explain_ltxt=str(e))
            if inbound_message.receipt.thread_id:
                error_result.assign_thread_id(inbound_message.receipt.thread_id)
            message = None

        trace_event(
            self.profile.settings,
            message,
            outcome="Dispatcher.handle_message.START",
        )

        context = RequestContext(profile)
        context.message = message
        context.message_receipt = inbound_message.receipt

        responder = DispatcherResponder(
            context,
            inbound_message,
            send_outbound,
            send_webhook,
            reply_session_id=inbound_message.session_id,
            reply_to_verkey=inbound_message.receipt.sender_verkey,
        )

        context.injector.bind_instance(BaseResponder, responder)

        async with profile.session(context._context) as session:
            connection_mgr = ConnectionManager(session)
            connection = await connection_mgr.find_inbound_connection(
                inbound_message.receipt
            )
            del connection_mgr
        if connection:
            inbound_message.connection_id = connection.connection_id

        context.connection_ready = connection and connection.is_ready
        context.connection_record = connection
        responder.connection_id = connection and connection.connection_id

        if error_result:
            await responder.send_reply(error_result)
            return

        handler_cls = context.message.Handler
        handler = handler_cls().handle
        if self.collector:
            handler = self.collector.wrap_coro(handler, [handler.__qualname__])
        await handler(context, responder)

        trace_event(
            self.profile.settings,
            context.message,
            outcome="Dispatcher.handle_message.END",
            perf_counter=r_time,
        )

    async def make_message(self, parsed_msg: dict) -> AgentMessage:
        """
        Deserialize a message dict into the appropriate message instance.

        Given a dict describing a message, this method
        returns an instance of the related message class.

        Args:
            parsed_msg: The parsed message

        Returns:
            An instance of the corresponding message class for this message

        Raises:
            MessageParseError: If the message doesn't specify @type
            MessageParseError: If there is no message class registered to handle
            the given type

        """

        registry: ProtocolRegistry = self.profile.inject(ProtocolRegistry)
        message_type = parsed_msg.get("@type")

        if not message_type:
            raise MessageParseError("Message does not contain '@type' parameter")

        try:
            message_cls = registry.resolve_message_class(message_type)
        except ProtocolMinorVersionNotSupported as e:
            raise MessageParseError(f"Problem parsing message type. {e}")

        if not message_cls:
            raise MessageParseError(f"Unrecognized message type {message_type}")

        try:
            instance = message_cls.deserialize(parsed_msg)
        except BaseModelError as e:
            raise MessageParseError(f"Error deserializing message: {e}") from e

        return instance

    async def complete(self, timeout: float = 0.1):
        """Wait for pending tasks to complete."""
        await self.task_queue.complete(timeout=timeout)


class DispatcherResponder(BaseResponder):
    """Handle outgoing messages from message handlers."""

    def __init__(
        self,
        context: RequestContext,
        inbound_message: InboundMessage,
        send_outbound: Coroutine,
        send_webhook: Coroutine = None,
        **kwargs,
    ):
        """
        Initialize an instance of `DispatcherResponder`.

        Args:
            context: The request context of the incoming message
            inbound_message: The inbound message triggering this handler
            send_outbound: Async function to send outbound message
            send_webhook: Async function to dispatch a webhook

        """
        super().__init__(**kwargs)
        self._context = context
        self._inbound_message = inbound_message
        self._send = send_outbound
        self._webhook = send_webhook

    async def create_outbound(
        self, message: Union[AgentMessage, str, bytes], **kwargs
    ) -> OutboundMessage:
        """
        Create an OutboundMessage from a message body.

        Args:
            message: The message payload
        """
        if isinstance(message, AgentMessage) and self._context.settings.get(
            "timing.enabled"
        ):
            # Inject the timing decorator
            in_time = (
                self._context.message_receipt and self._context.message_receipt.in_time
            )
            if not message._decorators.get("timing"):
                message._decorators["timing"] = {
                    "in_time": in_time,
                    "out_time": datetime_now(),
                }
        return await super().create_outbound(message, **kwargs)

    async def send_outbound(self, message: OutboundMessage):
        """
        Send outbound message.

        Args:
            message: The `OutboundMessage` to be sent
        """
        await self._send(self._context.profile, message, self._inbound_message)

    async def send_webhook(self, topic: str, payload: dict):
        """
        Dispatch a webhook.

        Args:
            topic: the webhook topic identifier
            payload: the webhook payload value
        """
        if self._webhook:
            await self._webhook(self._context.profile, topic, payload)



"""Websockets Transport classes and functions."""

import asyncio
import logging

from aiohttp import web, WSMessage, WSMsgType

from ...messaging.error import MessageParseError

from ..error import WireFormatParseError

from .base import BaseInboundTransport, InboundTransportSetupError

LOGGER = logging.getLogger(__name__)


class WsTransport(BaseInboundTransport):
    """Websockets Transport class."""

    def __init__(self, host: str, port: int, create_session, **kwargs) -> None:
        """
        Initialize an inbound WebSocket transport instance.

        Args:
            host: Host to listen on
            port: Port to listen on
            create_session: Method to create a new inbound session

        """
        super().__init__("ws", create_session, **kwargs)
        self.host = host
        self.port = port
        self.site: web.BaseSite = None

        # TODO: set scheme dynamically based on SSL settings (ws/wss)

    @property
    def scheme(self):
        """Accessor for this transport's scheme."""
        return self._scheme

    async def make_application(self) -> web.Application:
        """Construct the aiohttp application."""
        app = web.Application()
        app.add_routes([web.get("/", self.inbound_message_handler)])
        return app

    async def start(self) -> None:
        """
        Start this transport.

        Raises:
            InboundTransportSetupError: If there was an error starting the webserver

        """
        app = await self.make_application()
        runner = web.AppRunner(app)
        await runner.setup()
        self.site = web.TCPSite(runner, host=self.host, port=self.port)
        try:
            await self.site.start()
        except OSError:
            raise InboundTransportSetupError(
                "Unable to start websocket server with host "
                + f"'{self.host}' and port '{self.port}'\n"
            )

    async def stop(self) -> None:
        """Stop this transport."""
        if self.site:
            await self.site.stop()
            self.site = None

    async def inbound_message_handler(self, request):
        """
        Message handler for inbound messages.

        Args:
            request: aiohttp request object

        Returns:
            The web response

        """

        ws = web.WebSocketResponse()
        await ws.prepare(request)
        loop = asyncio.get_event_loop()

        client_info = {"host": request.host, "remote": request.remote}

        session = await self.create_session(
            accept_undelivered=True, can_respond=True, client_info=client_info
        )

        async with session:
            inbound = loop.create_task(ws.receive())
            outbound = loop.create_task(session.wait_response())

            while not ws.closed:
                await asyncio.wait(
                    (inbound, outbound), return_when=asyncio.FIRST_COMPLETED
                )

                if inbound.done():
                    msg: WSMessage = inbound.result()
                    LOGGER.info("Websocket received message: %s", msg.data)
                    if msg.type in (WSMsgType.TEXT, WSMsgType.BINARY):
                        try:
                            await session.receive(msg.data)
                        except (MessageParseError, WireFormatParseError):
                            await ws.close(1003)  # unsupported data error
                    elif msg.type == WSMsgType.ERROR:
                        LOGGER.error(
                            "Websocket connection closed with exception: %s",
                            ws.exception(),
                        )
                    if not ws.closed:
                        inbound = loop.create_task(ws.receive())

                if outbound.done() and not ws.closed:
                    # response would be None if session was closed
                    response = outbound.result()
                    if isinstance(response, bytes):
                        await ws.send_bytes(response)
                    else:
                        await ws.send_str(response)
                    session.clear_response()
                    outbound = loop.create_task(session.wait_response())

        if inbound and not inbound.done():
            inbound.cancel()
        if outbound and not outbound.done():
            outbound.cancel()

        if not ws.closed:
            await ws.close()
        LOGGER.info("Websocket connection closed")

        return ws



import json

from datetime import datetime, timezone
from unittest import TestCase

from .....didcomm_prefix import DIDCommPrefix

from ...message_types import PRESENTATION_ACK

from ..presentation_ack import PresentationAck


class TestPresentationAck(TestCase):
    """Presentation ack tests."""

    def test_init(self):
        """Test initializer."""
        pres_ack = PresentationAck()
        assert pres_ack.status == "OK"

    def test_type(self):
        """Test type."""
        pres_ack = PresentationAck()
        assert pres_ack._type == DIDCommPrefix.qualify_current(PRESENTATION_ACK)

    def test_deserialize(self):
        """Test deserialization."""
        dump = json.dumps(
            {"@type": DIDCommPrefix.qualify_current(PRESENTATION_ACK), "status": "OK"}
        )

        pres_ack = PresentationAck.deserialize(dump)
        assert type(pres_ack) == PresentationAck

    def test_serialize(self):
        """Test serialization."""
        pres_ack_dict = PresentationAck().serialize()
        pres_ack_dict.pop("@id")

        assert pres_ack_dict == {
            "@type": DIDCommPrefix.qualify_current(PRESENTATION_ACK),
            "status": "OK",
        }


class TestPresentationAckSchema(TestCase):
    """Test presentation ack schema"""

    def test_make_model(self):
        """Test making model."""
        pres_ack_dict = PresentationAck().serialize()
        """
        Looks like: {
            "@type": ".../present-proof/1.0/ack",
            "@id": "f49773e3-bd56-4868-a5f1-456d1e6d1a16",
            "status": "OK"
        }
        """

        model_instance = PresentationAck.deserialize(pres_ack_dict)
        assert isinstance(model_instance, PresentationAck)



import json
from datetime import datetime, timezone
from unittest import TestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.util import str_to_datetime, str_to_epoch

from .....didcomm_prefix import DIDCommPrefix

from ...message_types import ATTACH_DECO_IDS, PRESENTATION_PREVIEW, PRESENTATION_REQUEST

from ..presentation_request import PresentationRequest, PresentationRequestSchema


NOW_8601 = datetime.utcnow().replace(tzinfo=timezone.utc).isoformat(" ", "seconds")
NOW_EPOCH = str_to_epoch(NOW_8601)
CD_ID = "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag"
INDY_PROOF_REQ = json.loads(
    f"""{{
    "name": "proof-req",
    "version": "1.0",
    "nonce": "12345",
    "requested_attributes": {{
        "0_player_uuid": {{
            "name": "player",
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID}"
                }}
            ],
            "non_revoked": {{
                "from": {NOW_EPOCH},
                "to": {NOW_EPOCH}
            }}
        }},
        "0_screencapture_uuid": {{
            "name": "screenCapture",
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID}"
                }}
            ],
            "non_revoked": {{
                "from": {NOW_EPOCH},
                "to": {NOW_EPOCH}
            }}
        }}
    }},
    "requested_predicates": {{
        "0_highscore_GE_uuid": {{
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID}"
                }}
            ],
            "non_revoked": {{
                "from": {NOW_EPOCH},
                "to": {NOW_EPOCH}
            }}
        }}
    }}
}}"""
)

PRES_REQ = PresentationRequest(
    comment="Test",
    request_presentations_attach=[
        AttachDecorator.data_base64(
            mapping=INDY_PROOF_REQ,
            ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
        )
    ],
)


class TestPresentationRequest(TestCase):
    """Presentation request tests."""

    def test_init(self):
        """Test initializer."""
        assert PRES_REQ.request_presentations_attach[0].content == INDY_PROOF_REQ
        assert PRES_REQ.indy_proof_request(0) == INDY_PROOF_REQ

    def test_type(self):
        """Test type."""
        assert PRES_REQ._type == DIDCommPrefix.qualify_current(PRESENTATION_REQUEST)

    def test_deserialize(self):
        """Test deserialization."""
        dump = json.dumps(
            {
                "@type": DIDCommPrefix.qualify_current(PRESENTATION_REQUEST),
                "comment": "Hello World",
                "request_presentations~attach": [
                    AttachDecorator.data_base64(
                        mapping=INDY_PROOF_REQ,
                        ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
                    ).serialize()
                ],
            }
        )

        presentation_request = PresentationRequest.deserialize(dump)
        assert type(presentation_request) == PresentationRequest

    def test_serialize(self):
        """Test serialization."""
        pres_req_dict = PRES_REQ.serialize()
        pres_req_dict.pop("@id")

        assert pres_req_dict == {
            "@type": DIDCommPrefix.qualify_current(PRESENTATION_REQUEST),
            "request_presentations~attach": [
                AttachDecorator.data_base64(
                    mapping=INDY_PROOF_REQ,
                    ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
                ).serialize()
            ],
            "comment": "Test",
        }


class TestPresentationRequestSchema(TestCase):
    """Test presentation request schema"""

    def test_make_model(self):
        """Test making model."""
        pres_req_dict = PRES_REQ.serialize()
        """
        Looks like: {
            "@type": ".../present-proof/1.0/request-presentation",
            "@id": "f49773e3-bd56-4868-a5f1-456d1e6d1a16",
            "comment": "Test",
            "request_presentations~attach": [
                {
                    "mime-type": "application/json",
                    "data": {
                        "base64": "eyJuYW..."
                    }
                }
            ]
        }
        """

        model_instance = PRES_REQ.deserialize(pres_req_dict)
        assert isinstance(model_instance, PresentationRequest)



"""Multitenant admin routes."""

from marshmallow import fields, validate, validates_schema, ValidationError
from aiohttp import web
from aiohttp_apispec import (
    docs,
    request_schema,
    match_info_schema,
    response_schema,
    querystring_schema,
)

from ...admin.request_context import AdminRequestContext
from ...messaging.valid import JSONWebToken, UUIDFour
from ...messaging.models.base import BaseModelError
from ...messaging.models.openapi import OpenAPISchema
from ...storage.error import StorageError, StorageNotFoundError
from ...wallet.models.wallet_record import WalletRecord, WalletRecordSchema
from ...wallet.error import WalletSettingsError
from ...core.error import BaseError
from ...core.profile import ProfileManagerProvider
from ..manager import MultitenantManager
from ..error import WalletKeyMissingError


def format_wallet_record(wallet_record: WalletRecord):
    """Serialize a WalletRecord object."""

    wallet_info = wallet_record.serialize()

    # Hide wallet wallet key
    if "wallet.key" in wallet_info["settings"]:
        del wallet_info["settings"]["wallet.key"]

    return wallet_info


class MultitenantModuleResponseSchema(OpenAPISchema):
    """Response schema for multitenant module."""


class WalletIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking wallet id."""

    wallet_id = fields.Str(
        description="Subwallet identifier", required=True, example=UUIDFour.EXAMPLE
    )


class CreateWalletRequestSchema(OpenAPISchema):
    """Request schema for adding a new wallet which will be registered by the agent."""

    wallet_name = fields.Str(description="Wallet name", example="MyNewWallet")

    wallet_key = fields.Str(
        description="Master key used for key derivation.", example="MySecretKey123"
    )

    wallet_type = fields.Str(
        description="Type of the wallet to create",
        example="indy",
        default="in_memory",
        validate=validate.OneOf(
            [wallet_type for wallet_type in ProfileManagerProvider.MANAGER_TYPES]
        ),
    )

    wallet_dispatch_type = fields.Str(
        description="Webhook target dispatch type for this wallet. \
            default - Dispatch only to webhooks associated with this wallet. \
            base - Dispatch only to webhooks associated with the base wallet. \
            both - Dispatch to both webhook targets.",
        example="default",
        default="default",
        validate=validate.OneOf(["default", "both", "base"]),
    )

    wallet_webhook_urls = fields.List(
        fields.Str(
            description="Optional webhook URL to receive webhook messages",
            example="http://localhost:8022/webhooks",
        ),
        required=False,
        description="List of Webhook URLs associated with this subwallet",
    )

    label = fields.Str(
        description="Label for this wallet. This label is publicized\
            (self-attested) to other agents as part of forming a connection.",
        example="Alice",
    )

    image_url = fields.Str(
        description="Image url for this wallet. This image url is publicized\
            (self-attested) to other agents as part of forming a connection.",
        example="https://aries.ca/images/sample.png",
    )

    key_management_mode = fields.Str(
        description="Key management method to use for this wallet.",
        example=WalletRecord.MODE_MANAGED,
        default=WalletRecord.MODE_MANAGED,
        # MTODO: add unmanaged mode once implemented
        validate=validate.OneOf((WalletRecord.MODE_MANAGED,)),
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """
        Validate schema fields.

        Args:
            data: The data to validate

        Raises:
            ValidationError: If any of the fields do not validate

        """

        if data.get("wallet_type") == "indy":
            for field in ("wallet_key", "wallet_name"):
                if field not in data:
                    raise ValidationError("Missing required field", field)


class UpdateWalletRequestSchema(OpenAPISchema):
    """Request schema for updating a existing wallet."""

    wallet_dispatch_type = fields.Str(
        description="Webhook target dispatch type for this wallet. \
            default - Dispatch only to webhooks associated with this wallet. \
            base - Dispatch only to webhooks associated with the base wallet. \
            both - Dispatch to both webhook targets.",
        example="default",
        default="default",
        validate=validate.OneOf(["default", "both", "base"]),
    )

    wallet_webhook_urls = fields.List(
        fields.Str(
            description="Optional webhook URL to receive webhook messages",
            example="http://localhost:8022/webhooks",
        ),
        required=False,
        description="List of Webhook URLs associated with this subwallet",
    )

    label = fields.Str(
        description="Label for this wallet. This label is publicized\
            (self-attested) to other agents as part of forming a connection.",
        example="Alice",
    )

    image_url = fields.Str(
        description="Image url for this wallet. This image url is publicized\
            (self-attested) to other agents as part of forming a connection.",
        example="https://aries.ca/images/sample.png",
    )


class CreateWalletResponseSchema(WalletRecordSchema):
    """Response schema for creating a wallet."""

    token = fields.Str(
        description="Authorization token to authenticate wallet requests",
        example=JSONWebToken.EXAMPLE,
    )


class RemoveWalletRequestSchema(OpenAPISchema):
    """Request schema for removing a wallet."""

    wallet_key = fields.Str(
        description="Master key used for key derivation. Only required for \
            unmanaged wallets.",
        example="MySecretKey123",
    )


class CreateWalletTokenRequestSchema(OpenAPISchema):
    """Request schema for creating a wallet token."""

    wallet_key = fields.Str(
        description="Master key used for key derivation. Only required for \
            unamanged wallets.",
        example="MySecretKey123",
    )


class CreateWalletTokenResponseSchema(OpenAPISchema):
    """Response schema for creating a wallet token."""

    token = fields.Str(
        description="Authorization token to authenticate wallet requests",
        example=JSONWebToken.EXAMPLE,
    )


class WalletListSchema(OpenAPISchema):
    """Result schema for wallet list."""

    results = fields.List(
        fields.Nested(WalletRecordSchema()),
        description="List of wallet records",
    )


class WalletListQueryStringSchema(OpenAPISchema):
    """Parameters and validators for wallet list request query string."""

    wallet_name = fields.Str(description="Wallet name", example="MyNewWallet")


@docs(tags=["multitenancy"], summary="Query subwallets")
@querystring_schema(WalletListQueryStringSchema())
@response_schema(WalletListSchema(), 200, description="")
async def wallets_list(request: web.BaseRequest):
    """
    Request handler for listing all internal subwallets.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]

    query = {}
    wallet_name = request.query.get("wallet_name")
    if wallet_name:
        query["wallet_name"] = wallet_name

    async with context.session() as session:
        try:
            records = await WalletRecord.query(session, tag_filter=query)
            results = [format_wallet_record(record) for record in records]
            results.sort(key=lambda w: w["created_at"])
        except (StorageError, BaseModelError) as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(tags=["multitenancy"], summary="Get a single subwallet")
@match_info_schema(WalletIdMatchInfoSchema())
@response_schema(WalletRecordSchema(), 200, description="")
async def wallet_get(request: web.BaseRequest):
    """
    Request handler for getting a single subwallet.

    Args:
        request: aiohttp request object

    Raises:
        HTTPNotFound: if wallet_id does not match any known wallets

    """

    context: AdminRequestContext = request["context"]
    wallet_id = request.match_info["wallet_id"]

    async with context.session() as session:
        try:
            wallet_record = await WalletRecord.retrieve_by_id(session, wallet_id)
            result = format_wallet_record(wallet_record)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(tags=["multitenancy"], summary="Create a subwallet")
@request_schema(CreateWalletRequestSchema)
@response_schema(CreateWalletResponseSchema(), 200, description="")
async def wallet_create(request: web.BaseRequest):
    """
    Request handler for adding a new subwallet for handling by the agent.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]
    body = await request.json()

    key_management_mode = body.get("key_management_mode") or WalletRecord.MODE_MANAGED
    wallet_key = body.get("wallet_key")
    wallet_webhook_urls = body.get("wallet_webhook_urls") or []
    wallet_dispatch_type = body.get("wallet_dispatch_type") or "default"
    # If no webhooks specified, then dispatch only to base webhook targets
    if wallet_webhook_urls == []:
        wallet_dispatch_type = "base"

    settings = {
        "wallet.type": body.get("wallet_type") or "in_memory",
        "wallet.name": body.get("wallet_name"),
        "wallet.key": wallet_key,
        "wallet.webhook_urls": wallet_webhook_urls,
        "wallet.dispatch_type": wallet_dispatch_type,
    }

    label = body.get("label")
    image_url = body.get("image_url")
    if label:
        settings["default_label"] = label
    if image_url:
        settings["image_url"] = image_url

    async with context.session() as session:
        try:
            multitenant_mgr = session.inject(MultitenantManager)

            wallet_record = await multitenant_mgr.create_wallet(
                settings, key_management_mode
            )

            token = multitenant_mgr.create_auth_token(wallet_record, wallet_key)
        except BaseError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    result = {
        **format_wallet_record(wallet_record),
        "token": token,
    }
    return web.json_response(result)


@docs(tags=["multitenancy"], summary="Update a subwallet")
@match_info_schema(WalletIdMatchInfoSchema())
@request_schema(UpdateWalletRequestSchema)
@response_schema(WalletRecordSchema(), 200, description="")
async def wallet_update(request: web.BaseRequest):
    """
    Request handler for updating a existing subwallet for handling by the agent.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]
    wallet_id = request.match_info["wallet_id"]

    body = await request.json()
    wallet_webhook_urls = body.get("wallet_webhook_urls")
    wallet_dispatch_type = body.get("wallet_dispatch_type")
    label = body.get("label")
    image_url = body.get("image_url")

    if all(
        v is None for v in (wallet_webhook_urls, wallet_dispatch_type, label, image_url)
    ):
        raise web.HTTPBadRequest(reason="At least one parameter is required.")

    # adjust wallet_dispatch_type according to wallet_webhook_urls
    if wallet_webhook_urls and wallet_dispatch_type is None:
        wallet_dispatch_type = "default"
    if wallet_webhook_urls == []:
        wallet_dispatch_type = "base"

    # only parameters that are not none are updated
    settings = {}
    if wallet_webhook_urls is not None:
        settings["wallet.webhook_urls"] = wallet_webhook_urls
    if wallet_dispatch_type is not None:
        settings["wallet.dispatch_type"] = wallet_dispatch_type
    if label is not None:
        settings["default_label"] = label
    if image_url is not None:
        settings["image_url"] = image_url

    async with context.session() as session:
        try:
            multitenant_mgr = session.inject(MultitenantManager)
            wallet_record = await multitenant_mgr.update_wallet(wallet_id, settings)
            result = format_wallet_record(wallet_record)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except WalletSettingsError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(tags=["multitenancy"], summary="Get auth token for a subwallet")
@request_schema(CreateWalletTokenRequestSchema)
@response_schema(CreateWalletTokenResponseSchema(), 200, description="")
async def wallet_create_token(request: web.BaseRequest):
    """
    Request handler for creating an authorization token for a specific subwallet.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]
    wallet_id = request.match_info["wallet_id"]
    wallet_key = None

    if request.has_body:
        body = await request.json()
        wallet_key = body.get("wallet_key")

    async with context.session() as session:
        try:
            multitenant_mgr = session.inject(MultitenantManager)
            wallet_record = await WalletRecord.retrieve_by_id(session, wallet_id)

            if (not wallet_record.requires_external_key) and wallet_key:
                raise web.HTTPBadRequest(
                    reason=f"Wallet {wallet_id} doesn't require"
                    " the wallet key to be provided"
                )

            token = multitenant_mgr.create_auth_token(wallet_record, wallet_key)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except WalletKeyMissingError as err:
            raise web.HTTPUnauthorized(reason=err.roll_up) from err

    return web.json_response({"token": token})


@docs(
    tags=["multitenancy"],
    summary="Remove a subwallet",
)
@match_info_schema(WalletIdMatchInfoSchema())
@request_schema(RemoveWalletRequestSchema)
@response_schema(MultitenantModuleResponseSchema(), 200, description="")
async def wallet_remove(request: web.BaseRequest):
    """
    Request handler to remove a subwallet from agent and storage.

    Args:
        request: aiohttp request object.

    """

    context: AdminRequestContext = request["context"]
    wallet_id = request.match_info["wallet_id"]
    wallet_key = None

    if request.has_body:
        body = await request.json()
        wallet_key = body.get("wallet_key")

    async with context.session() as session:
        try:
            multitenant_mgr = session.inject(MultitenantManager)
            wallet_record = await WalletRecord.retrieve_by_id(session, wallet_id)

            if (not wallet_record.requires_external_key) and wallet_key:
                raise web.HTTPBadRequest(
                    reason=f"Wallet {wallet_id} doesn't require"
                    " the wallet key to be provided"
                )

            await multitenant_mgr.remove_wallet(wallet_id, wallet_key)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except WalletKeyMissingError as err:
            raise web.HTTPUnauthorized(reason=err.roll_up) from err

    return web.json_response({})


# MTODO: add wallet import route
# MTODO: add wallet export route
# MTODO: add rotate wallet key route


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/multitenancy/wallets", wallets_list, allow_head=False),
            web.post("/multitenancy/wallet", wallet_create),
            web.get("/multitenancy/wallet/{wallet_id}", wallet_get, allow_head=False),
            web.put("/multitenancy/wallet/{wallet_id}", wallet_update),
            web.post("/multitenancy/wallet/{wallet_id}/token", wallet_create_token),
            web.post("/multitenancy/wallet/{wallet_id}/remove", wallet_remove),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {"name": "multitenancy", "description": "Multitenant wallet management"}
    )



from tempfile import NamedTemporaryFile

from asynctest import TestCase as AsyncTestCase

from ...cache.base import BaseCache
from ...core.profile import ProfileManager
from ...core.protocol_registry import ProtocolRegistry
from ...transport.wire_format import BaseWireFormat

from ..default_context import DefaultContextBuilder
from ..injection_context import InjectionContext


class TestDefaultContext(AsyncTestCase):
    async def test_build_context(self):
        """Test context init."""

        builder = DefaultContextBuilder()
        result = await builder.build_context()
        assert isinstance(result, InjectionContext)

        for cls in (
            BaseCache,
            BaseWireFormat,
            ProfileManager,
            ProtocolRegistry,
        ):
            assert isinstance(result.inject(cls), cls)

        builder = DefaultContextBuilder(
            settings={
                "timing.enabled": True,
                "timing.log.file": NamedTemporaryFile().name,
                "multitenant.admin_enabled": True,
            }
        )
        result = await builder.build_context()
        assert isinstance(result, InjectionContext)



"""Test mediate grant message."""
from unittest import TestCase

from ...message_types import MEDIATE_GRANT
from ..mediate_grant import MediationGrant, MediationGrantSchema
from . import MessageTest


class TestMediateGrant(MessageTest, TestCase):
    """Test mediate grant message."""

    TYPE = MEDIATE_GRANT
    CLASS = MediationGrant
    SCHEMA = MediationGrantSchema
    VALUES = {"endpoint": "http://localhost:3000", "routing_keys": ["test_routing_key"]}



"""Store state for Mediation requests."""

from typing import Sequence

from marshmallow import EXCLUDE, fields

from .....core.profile import ProfileSession
from .....messaging.models.base_record import BaseRecord, BaseRecordSchema
from .....messaging.valid import INDY_RAW_PUBLIC_KEY
from .....storage.base import StorageDuplicateError, StorageNotFoundError


class MediationRecord(BaseRecord):
    """Class representing stored mediation information."""

    class Meta:
        """RouteRecord metadata."""

        schema_class = "MediationRecordSchema"

    RECORD_TYPE = "mediation_requests"
    RECORD_ID_NAME = "mediation_id"
    TAG_NAMES = {"state", "role", "connection_id"}

    STATE_REQUEST = "request"
    STATE_GRANTED = "granted"
    STATE_DENIED = "denied"

    ROLE_CLIENT = "client"
    ROLE_SERVER = "server"

    def __init__(
        self,
        *,
        mediation_id: str = None,
        state: str = None,
        role: str = None,
        connection_id: str = None,
        mediator_terms: Sequence[str] = None,
        recipient_terms: Sequence[str] = None,
        routing_keys: Sequence[str] = None,
        endpoint: str = None,
        **kwargs,
    ):
        """__init__.

        Args:
            mediation_id (str): (Optional) manually set record ID
            state (str): state, defaults to 'request_received'
            role (str): role in mediation, defaults to 'server'
            connection_id (str): ID of connection requesting or managing mediation
            mediator_terms (Sequence[str]): mediator_terms
            recipient_terms (Sequence[str]): recipient_terms
            routing_keys (Sequence[str]): keys in mediator control used to
            receive incoming messages
            endpoint (str): mediators endpoint
            kwargs: Pass arguments through to BaseRecord.__init__

        """
        super().__init__(mediation_id, state or self.STATE_REQUEST, **kwargs)
        self.role = role if role else self.ROLE_SERVER
        self.connection_id = connection_id
        self.mediator_terms = list(mediator_terms) if mediator_terms else []
        self.recipient_terms = list(recipient_terms) if recipient_terms else []
        self.routing_keys = list(routing_keys) if routing_keys else []
        self.endpoint = endpoint

    def __eq__(self, other: "MediationRecord"):
        """Equality check."""
        if not isinstance(other, MediationRecord):
            return False
        return (
            self.mediation_id == other.mediation_id
            and self.record_tags == other.record_tags
            and self.record_value == other.record_value
        )

    @property
    def mediation_id(self) -> str:
        """Get Mediation ID."""
        return self._id

    @property
    def state(self) -> str:
        """Get Mediation state."""
        return self._state

    @state.setter
    def state(self, state):
        """Setter for state."""
        if state not in [
            MediationRecord.STATE_DENIED,
            MediationRecord.STATE_GRANTED,
            MediationRecord.STATE_REQUEST,
        ]:
            raise ValueError(
                f"{state} is not a valid state, "
                f"must be one of ("
                f"{MediationRecord.STATE_DENIED}, "
                f"{MediationRecord.STATE_GRANTED}, "
                f"{MediationRecord.STATE_REQUEST}"
            )
        self._state = state

    @property
    def record_value(self) -> dict:
        """Return values of record as dictionary."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "mediator_terms",
                "recipient_terms",
                "routing_keys",
                "endpoint",
            )
        }

    @classmethod
    async def retrieve_by_connection_id(
        cls, session: ProfileSession, connection_id: str
    ) -> "MediationRecord":
        """Retrieve a mediation record by connection ID.

        Args:
            session (ProfileSession): session
            connection_id (str): connection_id

        Returns:
            MediationRecord: retrieved record

        """
        tag_filter = {"connection_id": connection_id}
        return await cls.retrieve_by_tag_filter(session, tag_filter)

    @classmethod
    async def exists_for_connection_id(
        cls, session: ProfileSession, connection_id: str
    ) -> bool:
        """Return whether a mediation record exists for the given connection.

        Args:
            session (ProfileSession): session
            connection_id (str): connection_id

        Returns:
            bool: whether record exists

        """
        tag_filter = {"connection_id": connection_id}
        try:
            record = await cls.retrieve_by_tag_filter(session, tag_filter)
        except StorageNotFoundError:
            return False
        except StorageDuplicateError:
            return True

        return bool(record)


class MediationRecordSchema(BaseRecordSchema):
    """MediationRecordSchema schema."""

    class Meta:
        """MediationRecordSchema metadata."""

        model_class = MediationRecord
        unknown = EXCLUDE

    mediation_id = fields.Str(required=False)
    role = fields.Str(required=True)
    connection_id = fields.Str(required=True)
    mediator_terms = fields.List(fields.Str(), required=False)
    recipient_terms = fields.List(fields.Str(), required=False)
    routing_keys = fields.List(fields.Str(**INDY_RAW_PUBLIC_KEY), required=False)
    endpoint = fields.Str(required=False)



"""Classes to manage connections."""

import asyncio
import json
import logging

from typing import Mapping, Sequence, Optional

from aries_cloudagent.protocols.coordinate_mediation.v1_0.manager import (
    MediationManager,
)

from ....connections.base_manager import BaseConnectionManager
from ....connections.models.conn_record import ConnRecord
from ....connections.util import mediation_record_if_id
from ....core.error import BaseError
from ....core.profile import ProfileSession
from ....indy.holder import IndyHolder
from ....messaging.responder import BaseResponder
from ....messaging.decorators.attach_decorator import AttachDecorator
from ....ledger.base import BaseLedger
from ....ledger.error import LedgerError
from ....multitenant.manager import MultitenantManager
from ....storage.error import StorageNotFoundError
from ....transport.inbound.receipt import MessageReceipt
from ....wallet.base import BaseWallet
from ....wallet.util import naked_to_did_key, b64_to_bytes, did_key_to_naked

from ...connections.v1_0.manager import ConnectionManager
from ...connections.v1_0.messages.connection_invitation import ConnectionInvitation
from ...didcomm_prefix import DIDCommPrefix
from ...didexchange.v1_0.manager import DIDXManager
from ...issue_credential.v1_0.models.credential_exchange import V10CredentialExchange
from ...issue_credential.v2_0.messages.cred_offer import V20CredOffer
from ...issue_credential.v2_0.models.cred_ex_record import V20CredExRecord
from ...present_proof.v1_0.manager import PresentationManager
from ...present_proof.v1_0.message_types import PRESENTATION_REQUEST
from ...present_proof.v1_0.messages.presentation_proposal import PresentationProposal
from ...present_proof.v1_0.models.presentation_exchange import V10PresentationExchange
from ...present_proof.v1_0.util.indy import indy_proof_req_preview2indy_requested_creds

from .messages.invitation import HSProto, InvitationMessage
from .messages.problem_report import ProblemReport
from .messages.reuse import HandshakeReuse
from .messages.reuse_accept import HandshakeReuseAccept
from .messages.service import Service as ServiceMessage
from .models.invitation import InvitationRecord


class OutOfBandManagerError(BaseError):
    """Out of band error."""


class OutOfBandManagerNotImplementedError(BaseError):
    """Out of band error for unimplemented functionality."""


class OutOfBandManager(BaseConnectionManager):
    """Class for managing out of band messages."""

    def __init__(self, session: ProfileSession):
        """
        Initialize a OutOfBandManager.

        Args:
            session: The profile session for this out of band manager
        """
        self._session = session
        self._logger = logging.getLogger(__name__)
        super().__init__(self._session)

    @property
    def session(self) -> ProfileSession:
        """
        Accessor for the current profile session.

        Returns:
            The profile session for this connection manager

        """
        return self._session

    async def create_invitation(
        self,
        my_label: str = None,
        my_endpoint: str = None,
        auto_accept: bool = None,
        public: bool = False,
        hs_protos: Sequence[HSProto] = None,
        multi_use: bool = False,
        alias: str = None,
        attachments: Sequence[Mapping] = None,
        metadata: dict = None,
        mediation_id: str = None,
    ) -> InvitationRecord:
        """
        Generate new connection invitation.

        This interaction represents an out-of-band communication channel. In the future
        and in practice, these sort of invitations will be received over any number of
        channels such as SMS, Email, QR Code, NFC, etc.

        Args:
            my_label: label for this connection
            my_endpoint: endpoint where other party can reach me
            auto_accept: auto-accept a corresponding connection request
                (None to use config)
            public: set to create an invitation from the public DID
            hs_protos: list of handshake protocols to include
            multi_use: set to True to create an invitation for multiple-use connection
            alias: optional alias to apply to connection for later use
            attachments: list of dicts in form of {"id": ..., "type": ...}

        Returns:
            Invitation record

        """
        mediation_mgr = MediationManager(self._session.profile)
        mediation_record = await mediation_record_if_id(
            self._session,
            mediation_id,
            or_default=True,
        )
        keylist_updates = None

        if not (hs_protos or attachments):
            raise OutOfBandManagerError(
                "Invitation must include handshake protocols, "
                "request attachments, or both"
            )

        wallet = self._session.inject(BaseWallet)

        # Multitenancy setup
        multitenant_mgr = self._session.inject(MultitenantManager, required=False)
        wallet_id = self._session.settings.get("wallet.id")

        accept = bool(
            auto_accept
            or (
                auto_accept is None
                and self._session.settings.get("debug.auto_accept_requests")
            )
        )
        if public:
            if multi_use:
                raise OutOfBandManagerError(
                    "Cannot create public invitation with multi_use"
                )
            if metadata:
                raise OutOfBandManagerError(
                    "Cannot store metadata on public invitations"
                )

        message_attachments = []
        for atch in attachments or []:
            a_type = atch.get("type")
            a_id = atch.get("id")

            if a_type == "credential-offer":
                try:
                    cred_ex_rec = await V10CredentialExchange.retrieve_by_id(
                        self._session,
                        a_id,
                    )
                    message_attachments.append(
                        InvitationMessage.wrap_message(
                            cred_ex_rec.credential_offer_dict
                        )
                    )
                except StorageNotFoundError:
                    cred_ex_rec = await V20CredExRecord.retrieve_by_id(
                        self._session,
                        a_id,
                    )
                    message_attachments.append(
                        InvitationMessage.wrap_message(
                            V20CredOffer.deserialize(
                                cred_ex_rec.cred_offer
                            ).offer()  # default to indy format: will change for DIF
                        )
                    )
            elif a_type == "present-proof":
                pres_ex_rec = await V10PresentationExchange.retrieve_by_id(
                    self._session,
                    a_id,
                )
                message_attachments.append(
                    InvitationMessage.wrap_message(
                        pres_ex_rec.presentation_request_dict
                    )
                )
            else:
                raise OutOfBandManagerError(f"Unknown attachment type: {a_type}")

        handshake_protocols = [
            DIDCommPrefix.qualify_current(hsp.name) for hsp in hs_protos or []
        ] or None
        if public:
            if not self._session.settings.get("public_invites"):
                raise OutOfBandManagerError("Public invitations are not enabled")

            public_did = await wallet.get_public_did()
            if not public_did:
                raise OutOfBandManagerError(
                    "Cannot create public invitation with no public DID"
                )

            invi_msg = InvitationMessage(  # create invitation message
                label=my_label or self._session.settings.get("default_label"),
                handshake_protocols=handshake_protocols,
                request_attach=message_attachments,
                service=[f"did:sov:{public_did.did}"],
            )
            keylist_updates = await mediation_mgr.add_key(
                public_did.verkey, keylist_updates
            )
            ledger = self._session.inject(BaseLedger)
            try:
                async with ledger:
                    base_url = await ledger.get_endpoint_for_did(public_did.did)
                    invi_url = invi_msg.to_url(base_url)
            except LedgerError as ledger_x:
                raise OutOfBandManagerError(
                    "Error getting endpoint for public DID "
                    f"{public_did.did}: {ledger_x}"
                )

            conn_rec = ConnRecord(  # create connection record
                invitation_key=public_did.verkey,
                invitation_msg_id=invi_msg._id,
                their_role=ConnRecord.Role.REQUESTER.rfc23,
                state=ConnRecord.State.INVITATION.rfc23,
                accept=ConnRecord.ACCEPT_AUTO if accept else ConnRecord.ACCEPT_MANUAL,
                alias=alias,
            )

            await conn_rec.save(self._session, reason="Created new invitation")
            await conn_rec.attach_invitation(self._session, invi_msg)

            if multitenant_mgr and wallet_id:  # add mapping for multitenant relay
                await multitenant_mgr.add_key(
                    wallet_id, public_did.verkey, skip_if_exists=True
                )

        else:
            invitation_mode = (
                ConnRecord.INVITATION_MODE_MULTI
                if multi_use
                else ConnRecord.INVITATION_MODE_ONCE
            )

            if not my_endpoint:
                my_endpoint = self._session.settings.get("default_endpoint")

            # Create and store new invitation key
            connection_key = await wallet.create_signing_key()
            keylist_updates = await mediation_mgr.add_key(
                connection_key.verkey, keylist_updates
            )
            # Add mapping for multitenant relay
            if multitenant_mgr and wallet_id:
                await multitenant_mgr.add_key(wallet_id, connection_key.verkey)

            # Create connection record
            conn_rec = ConnRecord(
                invitation_key=connection_key.verkey,
                their_role=ConnRecord.Role.REQUESTER.rfc23,
                state=ConnRecord.State.INVITATION.rfc23,
                accept=ConnRecord.ACCEPT_AUTO if accept else ConnRecord.ACCEPT_MANUAL,
                invitation_mode=invitation_mode,
                alias=alias,
            )
            await conn_rec.save(self._session, reason="Created new connection")

            routing_keys = []
            # The base wallet can act as a mediator for all tenants
            if multitenant_mgr and wallet_id:
                base_mediation_record = await multitenant_mgr.get_default_mediator()

                if base_mediation_record:
                    routing_keys = base_mediation_record.routing_keys
                    my_endpoint = base_mediation_record.endpoint

                    # If we use a mediator for the base wallet we don't
                    # need to register the key at the subwallet mediator
                    # because it only needs to know the key of the base mediator
                    # sub wallet mediator -> base wallet mediator -> agent
                    keylist_updates = None
            if mediation_record:
                routing_keys = [*routing_keys, *mediation_record.routing_keys]
                my_endpoint = mediation_record.endpoint

                # Save that this invitation was created with mediation
                await conn_rec.metadata_set(
                    self._session, "mediation", {"id": mediation_id}
                )

                if keylist_updates:
                    responder = self._session.inject(BaseResponder, required=False)
                    await responder.send(
                        keylist_updates, connection_id=mediation_record.connection_id
                    )
            routing_keys = [
                key if len(key.split(":")) == 3 else naked_to_did_key(key)
                for key in routing_keys
            ]
            # Create connection invitation message
            # Note: Need to split this into two stages to support inbound routing
            # of invitations
            # Would want to reuse create_did_document and convert the result
            invi_msg = InvitationMessage(
                label=my_label or self._session.settings.get("default_label"),
                handshake_protocols=handshake_protocols,
                request_attach=message_attachments,
                service=[
                    ServiceMessage(
                        _id="#inline",
                        _type="did-communication",
                        recipient_keys=[naked_to_did_key(connection_key.verkey)],
                        service_endpoint=my_endpoint,
                        routing_keys=routing_keys,
                    )
                ],
            )
            invi_url = invi_msg.to_url()

            # Update connection record
            conn_rec.invitation_msg_id = invi_msg._id
            await conn_rec.save(self._session, reason="Added Invitation")
            await conn_rec.attach_invitation(self._session, invi_msg)

            if metadata:
                for key, value in metadata.items():
                    await conn_rec.metadata_set(self._session, key, value)

        return InvitationRecord(  # for return via admin API, not storage
            state=InvitationRecord.STATE_INITIAL,
            invi_msg_id=invi_msg._id,
            invitation=invi_msg.serialize(),
            invitation_url=invi_url,
        )

    async def receive_invitation(
        self,
        invi_msg: InvitationMessage,
        use_existing_connection: bool = True,
        auto_accept: bool = None,
        alias: str = None,
        mediation_id: str = None,
    ) -> dict:
        """Receive an out of band invitation message."""

        ledger: BaseLedger = self._session.inject(BaseLedger)

        if mediation_id:
            try:
                await mediation_record_if_id(self._session, mediation_id)
            except StorageNotFoundError:
                mediation_id = None

        # There must be exactly 1 service entry
        if len(invi_msg.service_blocks) + len(invi_msg.service_dids) != 1:
            raise OutOfBandManagerError("service array must have exactly one element")

        if not (invi_msg.request_attach or invi_msg.handshake_protocols):
            raise OutOfBandManagerError(
                "Invitation must specify handshake_protocols, request_attach, or both"
            )
        # Get the single service item
        if len(invi_msg.service_blocks) >= 1:
            service = invi_msg.service_blocks[0]
            public_did = None
        else:
            # If it's in the did format, we need to convert to a full service block
            # An existing connection can only be reused based on a public DID
            # in an out-of-band message (RFC 0434).
            service_did = invi_msg.service_dids[0]
            async with ledger:
                verkey = await ledger.get_key_for_did(service_did)
                did_key = naked_to_did_key(verkey)
                endpoint = await ledger.get_endpoint_for_did(service_did)
            public_did = service_did.split(":")[-1]
            service = ServiceMessage.deserialize(
                {
                    "id": "#inline",
                    "type": "did-communication",
                    "recipientKeys": [did_key],
                    "routingKeys": [],
                    "serviceEndpoint": endpoint,
                }
            )

        unq_handshake_protos = [
            HSProto.get(hsp)
            for hsp in dict.fromkeys(
                [
                    DIDCommPrefix.unqualify(proto)
                    for proto in invi_msg.handshake_protocols
                ]
            )
        ]

        # Reuse Connection - only if started by an invitation with Public DID
        conn_rec = None
        if public_did is not None:  # invite has public DID: seek existing connection
            tag_filter = {}
            post_filter = {}
            # post_filter["state"] = ConnRecord.State.COMPLETED.rfc160
            post_filter["their_public_did"] = public_did
            conn_rec = await self.find_existing_connection(
                tag_filter=tag_filter, post_filter=post_filter
            )
        if conn_rec is not None:
            num_included_protocols = len(unq_handshake_protos)
            num_included_req_attachments = len(invi_msg.request_attach)
            # With handshake protocol, request attachment; use existing connection
            if (
                num_included_protocols >= 1
                and num_included_req_attachments == 0
                and use_existing_connection
            ):
                await self.create_handshake_reuse_message(
                    invi_msg=invi_msg,
                    conn_record=conn_rec,
                )
                try:
                    await asyncio.wait_for(
                        self.check_reuse_msg_state(
                            conn_rec=conn_rec,
                        ),
                        15,
                    )
                    await conn_rec.metadata_delete(
                        session=self._session, key="reuse_msg_id"
                    )
                    if (
                        await conn_rec.metadata_get(self._session, "reuse_msg_state")
                        == "not_accepted"
                    ):
                        conn_rec = None
                    else:
                        await conn_rec.metadata_delete(
                            session=self._session, key="reuse_msg_state"
                        )
                except asyncio.TimeoutError:
                    # If no reuse_accepted or problem_report message was received within
                    # the 15s timeout then a new connection to be created
                    await conn_rec.metadata_delete(
                        session=self._session, key="reuse_msg_id"
                    )
                    await conn_rec.metadata_delete(
                        session=self._session, key="reuse_msg_state"
                    )
                    conn_rec.state = ConnRecord.State.ABANDONED.rfc160
                    await conn_rec.save(self._session, reason="Sent connection request")
                    conn_rec = None
            # Inverse of the following cases
            # Handshake_Protocol not included
            # Request_Attachment included
            # Use_Existing_Connection Yes
            # Handshake_Protocol included
            # Request_Attachment included
            # Use_Existing_Connection Yes
            elif not (
                (
                    num_included_protocols == 0
                    and num_included_req_attachments >= 1
                    and use_existing_connection
                )
                or (
                    num_included_protocols >= 1
                    and num_included_req_attachments >= 1
                    and use_existing_connection
                )
            ):
                conn_rec = None
        if conn_rec is None:
            if not unq_handshake_protos:
                raise OutOfBandManagerError(
                    (
                        "No existing connection exists and "
                        "handshake_protocol is missing"
                    )
                )
            # Create a new connection
            for proto in unq_handshake_protos:
                if proto is HSProto.RFC23:
                    didx_mgr = DIDXManager(self._session)
                    conn_rec = await didx_mgr.receive_invitation(
                        invitation=invi_msg,
                        their_public_did=public_did,
                        auto_accept=auto_accept,
                        alias=alias,
                        mediation_id=mediation_id,
                    )
                elif proto is HSProto.RFC160:
                    service.recipient_keys = [
                        did_key_to_naked(key) for key in service.recipient_keys or []
                    ]
                    service.routing_keys = [
                        did_key_to_naked(key) for key in service.routing_keys
                    ] or []
                    connection_invitation = ConnectionInvitation.deserialize(
                        {
                            "@id": invi_msg._id,
                            "@type": DIDCommPrefix.qualify_current(proto.name),
                            "label": invi_msg.label,
                            "recipientKeys": service.recipient_keys,
                            "serviceEndpoint": service.service_endpoint,
                            "routingKeys": service.routing_keys,
                        }
                    )
                    conn_mgr = ConnectionManager(self._session)
                    conn_rec = await conn_mgr.receive_invitation(
                        invitation=connection_invitation,
                        their_public_did=public_did,
                        auto_accept=auto_accept,
                        alias=alias,
                        mediation_id=mediation_id,
                    )
                if conn_rec is not None:
                    break

        # Request Attach
        if len(invi_msg.request_attach) >= 1 and conn_rec is not None:
            req_attach = invi_msg.request_attach[0]
            if isinstance(req_attach, AttachDecorator):
                if req_attach.data is not None:
                    req_attach_type = req_attach.data.json["@type"]
                    if DIDCommPrefix.unqualify(req_attach_type) == PRESENTATION_REQUEST:
                        proof_present_mgr = PresentationManager(self._session)
                        indy_proof_request = json.loads(
                            b64_to_bytes(
                                req_attach.data.json["request_presentations~attach"][0][
                                    "data"
                                ]["base64"]
                            )
                        )
                        present_request_msg = req_attach.data.json
                        service_deco = {}
                        oob_invi_service = service.serialize()
                        service_deco["recipientKeys"] = oob_invi_service.get(
                            "recipientKeys"
                        )
                        service_deco["routingKeys"] = oob_invi_service.get(
                            "routingKeys"
                        )
                        service_deco["serviceEndpoint"] = oob_invi_service.get(
                            "serviceEndpoint"
                        )
                        present_request_msg["~service"] = service_deco
                        presentation_ex_record = V10PresentationExchange(
                            connection_id=conn_rec.connection_id,
                            thread_id=present_request_msg["@id"],
                            initiator=V10PresentationExchange.INITIATOR_EXTERNAL,
                            role=V10PresentationExchange.ROLE_PROVER,
                            presentation_request=indy_proof_request,
                            presentation_request_dict=present_request_msg,
                            auto_present=self._session.context.settings.get(
                                "debug.auto_respond_presentation_request"
                            ),
                            trace=(invi_msg._trace is not None),
                        )

                        presentation_ex_record.presentation_request = indy_proof_request
                        presentation_ex_record = (
                            await proof_present_mgr.receive_request(
                                presentation_ex_record
                            )
                        )
                        if presentation_ex_record.auto_present:
                            presentation_preview = None
                            if presentation_ex_record.presentation_proposal_dict:
                                exchange_pres_proposal = PresentationProposal.deserialize(
                                    presentation_ex_record.presentation_proposal_dict
                                )
                                presentation_preview = (
                                    exchange_pres_proposal.presentation_proposal
                                )

                            try:
                                req_creds = (
                                    await indy_proof_req_preview2indy_requested_creds(
                                        indy_proof_request,
                                        presentation_preview,
                                        holder=self._session.inject(IndyHolder),
                                    )
                                )
                            except ValueError as err:
                                self._logger.warning(f"{err}")
                                return

                            (
                                presentation_ex_record,
                                presentation_message,
                            ) = await proof_present_mgr.create_presentation(
                                presentation_exchange_record=presentation_ex_record,
                                requested_credentials=req_creds,
                                comment=(
                                    "auto-presented for proof request nonce={}".format(
                                        indy_proof_request["nonce"]
                                    )
                                ),
                            )
                            responder = self._session.inject(
                                BaseResponder, required=False
                            )
                            connection_targets = await self.fetch_connection_targets(
                                connection=conn_rec
                            )
                            if presentation_message is None:
                                raise OutOfBandManagerError(
                                    "No presentation for proof request nonce={}".format(
                                        indy_proof_request["nonce"]
                                    )
                                )
                            else:
                                if responder:
                                    await responder.send(
                                        message=presentation_message,
                                        target_list=connection_targets,
                                    )
                                return presentation_message.serialize()
                        else:
                            raise OutOfBandManagerError(
                                (
                                    "auto_present setting in configuration is"
                                    " set to false. Cannot respond automatically"
                                    " to presentation requests, building"
                                )
                            )
                    else:
                        raise OutOfBandManagerError(
                            (
                                "Unsupported request~attach type, "
                                "only request-presentation is supported"
                            )
                        )
            else:
                raise OutOfBandManagerError("request~attach is not properly formatted")
        else:
            return conn_rec.serialize()

    async def find_existing_connection(
        self,
        tag_filter: dict,
        post_filter: dict,
    ) -> Optional[ConnRecord]:
        """
        Find existing ConnRecord.

        Args:
            tag_filter: The filter dictionary to apply
            post_filter: Additional value filters to apply matching positively,
                with sequence values specifying alternatives to match (hit any)

        Returns:
            ConnRecord or None

        """
        conn_records = await ConnRecord.query(
            self._session,
            tag_filter=tag_filter,
            post_filter_positive=post_filter,
            alt=True,
        )
        if not conn_records:
            return None
        else:
            for conn_rec in conn_records:
                if conn_rec.state == "active":
                    return conn_rec
            return None

    async def check_reuse_msg_state(
        self,
        conn_rec: ConnRecord,
    ):
        """
        Check reuse message state from the ConnRecord Metadata.

        Args:
            conn_rec: The required ConnRecord with updated metadata

        Returns:

        """
        received = False
        while not received:
            if (
                not await conn_rec.metadata_get(self._session, "reuse_msg_state")
                == "initial"
            ):
                received = True
        return

    async def create_handshake_reuse_message(
        self,
        invi_msg: InvitationMessage,
        conn_record: ConnRecord,
    ) -> None:
        """
        Create and Send a Handshake Reuse message under RFC 0434.

        Args:
            invi_msg: OOB Invitation Message
            service: Service block extracted from the OOB invitation

        Returns:

        Raises:
            OutOfBandManagerError: If there is an issue creating or
            sending the OOB invitation

        """
        try:
            # ID of Out-of-Band invitation to use as a pthid
            # pthid = invi_msg._id
            pthid = conn_record.invitation_msg_id
            reuse_msg = HandshakeReuse()
            thid = reuse_msg._id
            reuse_msg.assign_thread_id(thid=thid, pthid=pthid)
            connection_targets = await self.fetch_connection_targets(
                connection=conn_record
            )
            responder = self._session.inject(BaseResponder, required=False)
            if responder:
                await responder.send(
                    message=reuse_msg,
                    target_list=connection_targets,
                )
                await conn_record.metadata_set(
                    session=self._session, key="reuse_msg_id", value=reuse_msg._id
                )
                await conn_record.metadata_set(
                    session=self._session, key="reuse_msg_state", value="initial"
                )
        except Exception as err:
            raise OutOfBandManagerError(
                f"Error on creating and sending a handshake reuse message: {err}"
            )

    async def receive_reuse_message(
        self,
        reuse_msg: HandshakeReuse,
        receipt: MessageReceipt,
    ) -> None:
        """
        Receive and process a HandshakeReuse message under RFC 0434.

        Process a `HandshakeReuse` message by looking up
        the connection records using the MessageReceipt sender DID.

        Args:
            reuse_msg: The `HandshakeReuse` to process
            receipt: The message receipt

        Returns:

        Raises:
            OutOfBandManagerError: If the existing connection is not active
            or the connection does not exists

        """
        try:
            invi_msg_id = reuse_msg._thread.pthid
            reuse_msg_id = reuse_msg._thread.thid
            tag_filter = {}
            post_filter = {}
            # post_filter["state"] = "active"
            # tag_filter["their_did"] = receipt.sender_did
            post_filter["invitation_msg_id"] = invi_msg_id
            conn_record = await self.find_existing_connection(
                tag_filter=tag_filter, post_filter=post_filter
            )
            responder = self._session.inject(BaseResponder, required=False)
            if conn_record is not None:
                # For ConnRecords created using did-exchange
                reuse_accept_msg = HandshakeReuseAccept()
                reuse_accept_msg.assign_thread_id(thid=reuse_msg_id, pthid=invi_msg_id)
                connection_targets = await self.fetch_connection_targets(
                    connection=conn_record
                )
                if responder:
                    await responder.send(
                        message=reuse_accept_msg,
                        target_list=connection_targets,
                    )
                # This is not required as now we attaching the invitation_msg_id
                # using original invitation [from existing connection]
                #
                # Delete the ConnRecord created; re-use existing connection
                # invi_id_post_filter = {}
                # invi_id_post_filter["invitation_msg_id"] = invi_msg_id
                # conn_rec_to_delete = await self.find_existing_connection(
                #     tag_filter={},
                #     post_filter=invi_id_post_filter,
                # )
                # if conn_rec_to_delete is not None:
                #     if conn_record.connection_id != conn_rec_to_delete.connection_id:
                #         await conn_rec_to_delete.delete_record(session=self._session)
            else:
                conn_record = await self.find_existing_connection(
                    tag_filter={"their_did": receipt.sender_did}, post_filter={}
                )
                # Problem Report is redundant in this case as with no active
                # connection, it cannot reach the invitee any way
                if conn_record is not None:
                    # For ConnRecords created using RFC 0160 connections
                    reuse_accept_msg = HandshakeReuseAccept()
                    reuse_accept_msg.assign_thread_id(
                        thid=reuse_msg_id, pthid=invi_msg_id
                    )
                    connection_targets = await self.fetch_connection_targets(
                        connection=conn_record
                    )
                    if responder:
                        await responder.send(
                            message=reuse_accept_msg,
                            target_list=connection_targets,
                        )
        except StorageNotFoundError:
            raise OutOfBandManagerError(
                (f"No existing ConnRecord found for OOB Invitee, {receipt.sender_did}"),
            )

    async def receive_reuse_accepted_message(
        self,
        reuse_accepted_msg: HandshakeReuseAccept,
        receipt: MessageReceipt,
        conn_record: ConnRecord,
    ) -> None:
        """
        Receive and process a HandshakeReuseAccept message under RFC 0434.

        Process a `HandshakeReuseAccept` message by updating the ConnRecord metadata
        state to `accepted`.

        Args:
            reuse_accepted_msg: The `HandshakeReuseAccept` to process
            receipt: The message receipt

        Returns:

        Raises:
            OutOfBandManagerError: if there is an error in processing the
            HandshakeReuseAccept message

        """
        try:
            invi_msg_id = reuse_accepted_msg._thread.pthid
            thread_reuse_msg_id = reuse_accepted_msg._thread.thid
            conn_reuse_msg_id = await conn_record.metadata_get(
                session=self._session, key="reuse_msg_id"
            )
            assert thread_reuse_msg_id == conn_reuse_msg_id
            await conn_record.metadata_set(
                session=self._session, key="reuse_msg_state", value="accepted"
            )
        except Exception as e:
            raise OutOfBandManagerError(
                (
                    (
                        "Error processing reuse accepted message "
                        f"for OOB invitation {invi_msg_id}, {e}"
                    )
                )
            )

    async def receive_problem_report(
        self,
        problem_report: ProblemReport,
        receipt: MessageReceipt,
        conn_record: ConnRecord,
    ) -> None:
        """
        Receive and process a ProblemReport message from the inviter to invitee.

        Process a `ProblemReport` message by updating  the ConnRecord metadata
        state to `not_accepted`.

        Args:
            problem_report: The `ProblemReport` to process
            receipt: The message receipt

        Returns:

        Raises:
            OutOfBandManagerError: if there is an error in processing the
            HandshakeReuseAccept message

        """
        try:
            invi_msg_id = problem_report._thread.pthid
            thread_reuse_msg_id = problem_report._thread.thid
            conn_reuse_msg_id = await conn_record.metadata_get(
                session=self._session, key="reuse_msg_id"
            )
            assert thread_reuse_msg_id == conn_reuse_msg_id
            await conn_record.metadata_set(
                session=self._session, key="reuse_msg_state", value="not_accepted"
            )
        except Exception as e:
            raise OutOfBandManagerError(
                (
                    (
                        "Error processing problem report message "
                        f"for OOB invitation {invi_msg_id}, {e}"
                    )
                )
            )



"""Classes for BaseStorage-based record management."""

import json
import sys
import uuid

from datetime import datetime
from typing import Any, Mapping, Sequence, Union

from marshmallow import fields

from ...cache.base import BaseCache
from ...config.settings import BaseSettings
from ...core.profile import ProfileSession
from ...storage.base import BaseStorage, StorageDuplicateError, StorageNotFoundError
from ...storage.record import StorageRecord

from .base import BaseModel, BaseModelSchema
from ..responder import BaseResponder
from ..util import datetime_to_str, time_now
from ..valid import INDY_ISO8601_DATETIME


def match_post_filter(
    record: dict,
    post_filter: dict,
    positive: bool = True,
    alt: bool = False,
) -> bool:
    """Determine if a record value matches the post-filter.

    Args:
        record: record to check
        post_filter: filter to apply (empty or None filter matches everything)
        positive: whether matching all filter criteria positively or negatively
        alt: set to match any (positive=True) value or miss all (positive=False)
            values in post_filter
    """
    if not post_filter:
        return True

    if alt:
        return (
            positive
            and all(
                record.get(k) and record.get(k) in alts
                for k, alts in post_filter.items()
            )
        ) or (
            (not positive)
            and all(
                record.get(k) and record.get(k) not in alts
                for k, alts in post_filter.items()
            )
        )

    for k, v in post_filter.items():
        if record.get(k) != v:
            return not positive

    return positive


class BaseRecord(BaseModel):
    """Represents a single storage record."""

    class Meta:
        """BaseRecord metadata."""

    DEFAULT_CACHE_TTL = 60
    RECORD_ID_NAME = "id"
    RECORD_TYPE = None
    WEBHOOK_TOPIC = None
    LOG_STATE_FLAG = None
    TAG_NAMES = {"state"}

    def __init__(
        self,
        id: str = None,
        state: str = None,
        *,
        created_at: Union[str, datetime] = None,
        updated_at: Union[str, datetime] = None,
    ):
        """Initialize a new BaseRecord."""
        if not self.RECORD_TYPE:
            raise TypeError(
                "Cannot instantiate abstract class {} with no RECORD_TYPE".format(
                    self.__class__.__name__
                )
            )
        self._id = id
        self._last_state = state
        self.state = state
        self.created_at = datetime_to_str(created_at)
        self.updated_at = datetime_to_str(updated_at)

    @classmethod
    def from_storage(cls, record_id: str, record: Mapping[str, Any]):
        """Initialize a record from its stored representation.

        Args:
            record_id: The unique record identifier
            record: The stored representation
        """
        record_id_name = cls.RECORD_ID_NAME
        if record_id_name in record:
            raise ValueError(f"Duplicate {record_id_name} inputs; {record}")
        params = dict(**record)
        params[record_id_name] = record_id
        return cls(**params)

    @classmethod
    def get_tag_map(cls) -> Mapping[str, str]:
        """Accessor for the set of defined tags."""
        return {tag.lstrip("~"): tag for tag in cls.TAG_NAMES or ()}

    @property
    def storage_record(self) -> StorageRecord:
        """Accessor for a `StorageRecord` representing this record."""
        return StorageRecord(
            self.RECORD_TYPE, json.dumps(self.value), self.tags, self._id
        )

    @property
    def record_value(self) -> dict:
        """Accessor to define custom properties for the JSON record value."""
        return {}

    @property
    def value(self) -> dict:
        """Accessor for the JSON record value generated for this record."""
        ret = self.strip_tag_prefix(self.tags)
        ret.update({"created_at": self.created_at, "updated_at": self.updated_at})
        ret.update(self.record_value)
        return ret

    @property
    def record_tags(self) -> dict:
        """Accessor to define implementation-specific tags."""
        return {
            tag: getattr(self, prop)
            for (prop, tag) in self.get_tag_map().items()
            if getattr(self, prop) is not None
        }

    @property
    def tags(self) -> dict:
        """Accessor for the record tags generated for this record."""
        tags = self.record_tags
        return tags

    @classmethod
    async def get_cached_key(cls, session: ProfileSession, cache_key: str):
        """Shortcut method to fetch a cached key value.

        Args:
            session: The profile session to use
            cache_key: The unique cache identifier
        """
        if not cache_key:
            return
        cache = session.inject(BaseCache, required=False)
        if cache:
            return await cache.get(cache_key)

    @classmethod
    async def set_cached_key(
        cls, session: ProfileSession, cache_key: str, value: Any, ttl=None
    ):
        """Shortcut method to set a cached key value.

        Args:
            session: The profile session to use
            cache_key: The unique cache identifier
            value: The value to cache
            ttl: The cache ttl
        """
        if not cache_key:
            return
        cache = session.inject(BaseCache, required=False)
        if cache:
            await cache.set(cache_key, value, ttl or cls.DEFAULT_CACHE_TTL)

    @classmethod
    async def clear_cached_key(cls, session: ProfileSession, cache_key: str):
        """Shortcut method to clear a cached key value, if any.

        Args:
            session: The profile session to use
            cache_key: The unique cache identifier
        """
        if not cache_key:
            return
        cache = session.inject(BaseCache, required=False)
        if cache:
            await cache.clear(cache_key)

    @classmethod
    async def retrieve_by_id(
        cls, session: ProfileSession, record_id: str
    ) -> "BaseRecord":
        """Retrieve a stored record by ID.

        Args:
            session: The profile session to use
            record_id: The ID of the record to find
        """
        storage = session.inject(BaseStorage)
        result = await storage.get_record(
            cls.RECORD_TYPE, record_id, {"retrieveTags": False}
        )
        vals = json.loads(result.value)
        return cls.from_storage(record_id, vals)

    @classmethod
    async def retrieve_by_tag_filter(
        cls, session: ProfileSession, tag_filter: dict, post_filter: dict = None
    ) -> "BaseRecord":
        """Retrieve a record by tag filter.

        Args:
            session: The profile session to use
            tag_filter: The filter dictionary to apply
            post_filter: Additional value filters to apply matching positively,
                with sequence values specifying alternatives to match (hit any)
        """
        storage = session.inject(BaseStorage)
        rows = await storage.find_all_records(
            cls.RECORD_TYPE,
            cls.prefix_tag_filter(tag_filter),
            options={"retrieveTags": False},
        )
        found = None
        for record in rows:
            vals = json.loads(record.value)
            if match_post_filter(vals, post_filter, alt=False):
                if found:
                    raise StorageDuplicateError(
                        "Multiple {} records located for {}{}".format(
                            cls.__name__,
                            tag_filter,
                            f", {post_filter}" if post_filter else "",
                        )
                    )
                found = cls.from_storage(record.id, vals)
        if not found:
            raise StorageNotFoundError(
                "{} record not found for {}{}".format(
                    cls.__name__, tag_filter, f", {post_filter}" if post_filter else ""
                )
            )
        return found

    @classmethod
    async def query(
        cls,
        session: ProfileSession,
        tag_filter: dict = None,
        *,
        post_filter_positive: dict = None,
        post_filter_negative: dict = None,
        alt: bool = False,
    ) -> Sequence["BaseRecord"]:
        """Query stored records.

        Args:
            session: The profile session to use
            tag_filter: An optional dictionary of tag filter clauses
            post_filter_positive: Additional value filters to apply matching positively
            post_filter_negative: Additional value filters to apply matching negatively
            alt: set to match any (positive=True) value or miss all (positive=False)
                values in post_filter
        """
        storage = session.inject(BaseStorage)
        rows = await storage.find_all_records(
            cls.RECORD_TYPE,
            cls.prefix_tag_filter(tag_filter),
            options={"retrieveTags": False},
        )
        result = []
        for record in rows:
            vals = json.loads(record.value)
            if match_post_filter(
                vals,
                post_filter_positive,
                positive=True,
                alt=alt,
            ) and match_post_filter(
                vals,
                post_filter_negative,
                positive=False,
                alt=alt,
            ):
                result.append(cls.from_storage(record.id, vals))
        return result

    async def save(
        self,
        session: ProfileSession,
        *,
        reason: str = None,
        log_params: Mapping[str, Any] = None,
        log_override: bool = False,
        webhook: bool = None,
    ) -> str:
        """Persist the record to storage.

        Args:
            session: The profile session to use
            reason: A reason to add to the log
            log_params: Additional parameters to log
            webhook: Flag to override whether the webhook is sent
        """
        new_record = None
        log_reason = reason or ("Updated record" if self._id else "Created record")
        try:
            self.updated_at = time_now()
            storage = session.inject(BaseStorage)
            if self._id:
                record = self.storage_record
                await storage.update_record(record, record.value, record.tags)
                new_record = False
            else:
                self._id = str(uuid.uuid4())
                self.created_at = self.updated_at
                await storage.add_record(self.storage_record)
                new_record = True
        finally:
            params = {self.RECORD_TYPE: self.serialize()}
            if log_params:
                params.update(log_params)
            if new_record is None:
                log_reason = f"FAILED: {log_reason}"
            self.log_state(
                log_reason, params, override=log_override, settings=session.settings
            )

        await self.post_save(session, new_record, self._last_state, webhook)
        self._last_state = self.state

        return self._id

    async def post_save(
        self,
        session: ProfileSession,
        new_record: bool,
        last_state: str,
        webhook: bool = None,
    ):
        """Perform post-save actions.

        Args:
            session: The profile session to use
            new_record: Flag indicating if the record was just created
            last_state: The previous state value
            webhook: Adjust whether the webhook is called
        """
        webhook_topic = self.webhook_topic
        if webhook is None:
            webhook = bool(webhook_topic) and (new_record or (last_state != self.state))
        if webhook:
            await self.send_webhook(
                session, self.webhook_payload, topic=self.webhook_topic
            )

    async def delete_record(self, session: ProfileSession):
        """Remove the stored record.

        Args:
            session: The profile session to use
        """
        if self._id:
            storage = session.inject(BaseStorage)
            await storage.delete_record(self.storage_record)
        # FIXME - update state and send webhook?

    @property
    def webhook_payload(self):
        """Return a JSON-serialized version of the record for the webhook."""
        return self.serialize()

    @property
    def webhook_topic(self):
        """Return the webhook topic value."""
        return self.WEBHOOK_TOPIC

    async def send_webhook(
        self, session: ProfileSession, payload: Any, topic: str = None
    ):
        """Send a standard webhook.

        Args:
            session: The profile session to use
            payload: The webhook payload
            topic: The webhook topic, defaulting to WEBHOOK_TOPIC
        """
        if not payload:
            return
        if not topic:
            topic = self.webhook_topic
            if not topic:
                return
        responder = session.inject(BaseResponder, required=False)
        if responder:
            await responder.send_webhook(topic, payload)

    @classmethod
    def log_state(
        cls,
        msg: str,
        params: dict = None,
        settings: BaseSettings = None,
        override: bool = False,
    ):
        """Print a message with increased visibility (for testing)."""
        if override or (
            cls.LOG_STATE_FLAG and settings and settings.get(cls.LOG_STATE_FLAG)
        ):
            out = msg + "\n"
            if params:
                for k, v in params.items():
                    out += f"    {k}: {v}\n"
            print(out, file=sys.stderr)

    @classmethod
    def strip_tag_prefix(cls, tags: dict):
        """Strip tilde from unencrypted tag names."""
        return (
            {(k[1:] if "~" in k else k): v for (k, v) in tags.items()} if tags else {}
        )

    @classmethod
    def prefix_tag_filter(cls, tag_filter: dict):
        """Prefix unencrypted tags used in the tag filter."""
        ret = None
        if tag_filter:
            tag_map = cls.get_tag_map()
            ret = {}
            for k, v in tag_filter.items():
                if k in ("$or", "$and") and isinstance(v, list):
                    ret[k] = [cls.prefix_tag_filter(clause) for clause in v]
                elif k == "$not" and isinstance(v, dict):
                    ret[k] = cls.prefix_tag_filter(v)
                else:
                    ret[tag_map.get(k, k)] = v
        return ret

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        if type(other) is type(self):
            return self.value == other.value and self.tags == other.tags
        return False


class BaseExchangeRecord(BaseRecord):
    """Represents a base record with event tracing capability."""

    def __init__(
        self,
        id: str = None,
        state: str = None,
        *,
        trace: bool = False,
        **kwargs,
    ):
        """Initialize a new BaseExchangeRecord."""
        super().__init__(id, state, **kwargs)
        self.trace = trace

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        if type(other) is type(self):
            return (
                self.value == other.value
                and self.tags == other.tags
                and self.trace == other.trace
            )
        return False


class BaseRecordSchema(BaseModelSchema):
    """Schema to allow serialization/deserialization of base records."""

    class Meta:
        """BaseRecordSchema metadata."""

        model_class = None

    state = fields.Str(
        required=False, description="Current record state", example="active"
    )
    created_at = fields.Str(
        required=False, description="Time of record creation", **INDY_ISO8601_DATETIME
    )
    updated_at = fields.Str(
        required=False,
        description="Time of last record update",
        **INDY_ISO8601_DATETIME,
    )


class BaseExchangeSchema(BaseRecordSchema):
    """Base schema for exchange records."""

    class Meta:
        """BaseExchangeSchema metadata."""

        model_class = BaseExchangeRecord

    trace = fields.Boolean(
        description="Record trace information, based on agent configuration",
        required=False,
        default=False,
    )
